#-*- coding:utf-8 -*-

# test_analysis.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Check that the examples work.
"""

import os
from os import environ
from os.path import dirname, abspath, isfile, join
import unittest

import pytest
from scipy.special import lambertw

import nngt

# set example dir
current_dir = dirname(abspath(__file__))
idx_testing = current_dir.find('testing')
example_dir = current_dir[:idx_testing] + 'doc/examples/'

# remove plotting and NEST
nngt.set_config("with_plot", False)
nngt.set_config("with_nest", False)

# set globals
glob = {"lambertw": lambertw}


# ---------- #
# Test class #
# ---------- #

@pytest.mark.mpi_skip
class TestExamples(unittest.TestCase):
    
    '''
    Class testing saving and loading functions.
    '''
    
    example_files = [
        example_dir + f for f in os.listdir(example_dir)
        if isfile(join(example_dir, f))
    ]

    @classmethod
    def tearDownClass(cls):
        try:
            os.remove("sp_graph.el")
        except:
            pass
    
    @property
    def test_name(self):
        return "test_examples"

    @unittest.skipIf(int(environ.get("OMP", 1)) == 1, 'Check only with OMP')
    def test_examples(self):
        '''
        Test that the example files execute correctly.
        '''
        for example in self.example_files:
            if example.endswith('.py'):
                with open(example) as f:
                    code = compile(f.read(), example, 'exec')
                    exec(code, glob)


# ---------- #
# Test suite #
# ---------- #

suite = unittest.TestLoader().loadTestsFromTestCase(TestExamples)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_attributes.py


"""
Test the validity of graph, node, and edge attributes as well as the
distribution generators.
"""

import os
import unittest

import numpy as np
import pytest

import nngt

from base_test import TestBasis, XmlHandler, network_dir
from tools_testing import foreach_graph


# ---------- #
# Test tools #
# ---------- #

def _results_theo(instruct):
    di_param = instruct["weights"]
    if (di_param["distribution"] == "uniform"
        or "corr" in di_param["distribution"]):
        return di_param["lower"], di_param["upper"]
    elif di_param["distribution"] == "gaussian":
        return di_param["avg"], di_param["std"]
    elif di_param["distribution"] == "lognormal":
        return di_param["position"], di_param["scale"]
    else:
        raise NotImplementedError("This distribution is not supported yet.")


def _results_exp(attrib, instruct):
    di_param = instruct["weights"]
    if (di_param["distribution"] == "uniform"
        or "corr" in di_param["distribution"]):
        return attrib.min(), attrib.max()
    elif di_param["distribution"] == "gaussian":
        return np.average(attrib), np.std(attrib)
    elif di_param["distribution"] == "lognormal":
        m = np.average(attrib)
        v = np.var(attrib)
        return np.log(m/np.sqrt(1+v/m**2)), np.sqrt(np.log(1+v/m**2))
    else:
        raise NotImplementedError("This distribution is not supported yet.")


# ---------- #
# Test class #
# ---------- #

class TestAttributes(TestBasis):
    
    '''
    Class testing the main methods of the :mod:`~nngt.generation` module.
    '''
    
    tolerance = 0.02
    
    @property
    def test_name(self):
        return "test_attributes"

    @unittest.skipIf(nngt.get_config('mpi'), 'Not checking for MPI')
    def gen_graph(self, graph_name):
        di_instructions = self.parser.get_graph_options(graph_name)
        graph = nngt.generate(di_instructions)
        graph.set_name(graph_name)
        return graph, di_instructions

    def test_node_attr(self):
        '''
        When generating graphs with weights, check that the expected properties
        are indeed obtained.
        '''
        g = nngt.Graph(100)
        ref_result = np.random.uniform(-1, 4, g.node_nb())
        g.set_node_attribute("nud", values=ref_result, value_type="double")
        computed_result = g.get_node_attributes(name="nud")
        self.assertTrue(np.allclose(ref_result, computed_result),
            '''Error on graph {}: unequal 'nud' attribute for tolerance {}.
            '''.format(g.name, self.tolerance))
    
    def test_nattr_default_values(self):
        g2 = nngt.Graph()

        # add a new node with attributes
        attributes = {
            'size': 2.,
            'color': 'blue',
            'a': 5,
            'blob': []
        }

        attribute_types = {
            'size': 'double',
            'color': 'string',
            'a': 'int',
            'blob': 'object'
        }

        g2.new_node(attributes=attributes, value_types=attribute_types)
        g2.new_node(2)
        g2.new_node(3, attributes={'size': [4., 5., 1.],
                    'color': ['r', 'g', 'b']},
                    value_types={'size': 'double', 'color': 'string'})

        # check all values
        # for the doubles:
        # NaN == NaN is false, so we need to check separately equality between
        # non-NaN entries and position of NaN entries
        double_res = np.array([2., np.NaN, np.NaN, 4., 5., 1.])
        isnan1     = np.isnan(g2.node_attributes['size'])
        isnan2     = np.isnan(double_res)
        self.assertTrue(np.all(isnan1 == isnan2))
        self.assertTrue(
            np.all(np.isclose(
                g2.node_attributes['size'][~isnan1], double_res[~isnan2]))
        )
        # for the others, just compare the lists
        self.assertEqual(
            g2.node_attributes['color'].tolist(),
            ['blue', '', '', 'r', 'g', 'b'])
        self.assertEqual(
            g2.node_attributes['a'].tolist(), [5, 0, 0, 0, 0, 0])
        self.assertEqual(
            g2.node_attributes['blob'].tolist(),
            [[], None, None, None, None, None])

    def test_user_defined(self):
        '''
        When generating graphs with weights, check that the expected properties
        are indeed obtained.
        '''
        avg = 50
        std = 6
        g = nngt.generation.gaussian_degree(avg, std, nodes=200)

        ref_result = np.random.uniform(0, 5, g.edge_nb())
        g.set_edge_attribute("ud", values=ref_result, value_type="double")

        computed_result = g.get_edge_attributes(name="ud")

        self.assertTrue(np.allclose(ref_result, computed_result),
            '''Error on graph {}: unequal 'ud' attribute for tolerance {}.
            '''.format(g.name, self.tolerance))

    def test_user_defined2(self):
        '''
        When generating graphs with weights, check that the expected properties
        are indeed obtained.
        '''
        avg = 50
        std = 6
        g = nngt.generation.gaussian_degree(avg, std, nodes=200)

        ref_result = np.full(g.edge_nb(), 4.)
        g.set_edge_attribute("ud2", val=4., value_type="double")

        computed_result = g.get_edge_attributes(name="ud2")
        self.assertTrue(np.allclose(ref_result, computed_result),
            '''Error on graph {}: unequal 'ud2' attribute for tolerance {}.
            '''.format(g.name, self.tolerance))

    @unittest.skipIf(nngt.get_config('mpi'), 'Not checking for MPI')
    def test_list_attributes(self):
        '''
        For list attributes, test that they are preserved as lists, and that
        some nodes or edges do not own references to the same list.
        '''
        avg = 25
        std = 3

        graph = nngt.generation.gaussian_degree(avg, std, nodes=1000)

        # --------------- #
        # node attributes #
        # --------------- #

        graph.new_node_attribute("nlist", value_type="object", val=[])

        nodes = [i for i in range(8, 49)]
        graph.set_node_attribute("nlist", val=[1], nodes=nodes)

        # update a fraction of the previously updated nodes
        nodes = [i for i in range(0, 41)]

        # to update the values, we need to get them to update the lists
        nlists = graph.get_node_attributes(name="nlist", nodes=nodes)

        for l in nlists:
            l.append(2)

        graph.set_node_attribute("nlist", values=nlists, nodes=nodes)

        # check that all lists are present
        nlists = graph.get_node_attributes(name="nlist")

        self.assertTrue(
            np.all(np.unique(nlists) == np.unique([[], [1], [2], [1, 2]])))

        # check that all nodes from 0 to 48 were updated
        self.assertTrue([] not in nlists[:49].tolist())

        # --------------- #
        # edge attributes #
        # --------------- #

        graph.new_edge_attribute("elist", value_type="object", val=[])

        nodes = list(range(8, 49))
        edges = graph.get_edges(source_node=nodes, target_node=nodes)
        graph.set_edge_attribute("elist", val=[1], edges=edges)

        # update a fraction of the previously updated nodes
        nodes  = list(range(0, 41))
        edges2 = graph.get_edges(source_node=nodes, target_node=nodes)

        # to update the values, we need to get them to update the lists
        elists = graph.get_edge_attributes(name="elist", edges=edges2)

        for l in elists:
            l.append(2)

        graph.set_edge_attribute("elist", values=elists, edges=edges2)

        # check that all lists are present
        elists = graph.get_edge_attributes(name="elist")

        self.assertTrue(
            np.all(np.unique(elists) == np.unique([[], [1], [2], [1, 2]])))

        # check that all edges where updated
        eattr1 = graph.get_edge_attributes(name="elist", edges=edges).tolist()
        eattr2 = graph.get_edge_attributes(name="elist", edges=edges2).tolist()

        self.assertTrue([] not in eattr1 and [] not in eattr2)

    @foreach_graph
    def test_weights(self, graph, instructions, **kwargs):
        '''
        When generating graphs with weights, check that the expected properties
        are indeed obtained.
        '''
        ref_result = _results_theo(instructions)

        weights    = graph.get_weights()
        computed_result = _results_exp(weights, instructions)

        self.assertTrue(np.allclose(
            ref_result, computed_result, self.tolerance),
            '''Error on graph {}: unequal weights for tolerance {}.
            '''.format(graph.name, self.tolerance))

    @foreach_graph
    def test_delays(self, graph, instructions, **kwargs):
        '''
        Test entirely run only if NEST is present on the computer.
        Check that delay distribution generated in NNGT, then in NEST, is
        conform to what was instructed.
        '''
        # get the informations from the weights
        di_distrib = instructions["weights"]
        distrib = di_distrib["distribution"]
        delays = graph.set_delays(distribution=distrib, parameters=di_distrib)
        ref_result = _results_theo(instructions)
        computed_result = _results_exp(delays, instructions)
        self.assertTrue(np.allclose(
            ref_result, computed_result, self.tolerance),
            '''Error on graph {}: unequal delays for tolerance {}.
            '''.format(graph.name, self.tolerance))
        # @todo
        #~ if nngt._config['with_nest']:
            #~ from nngt.simulation import make_nest_network
            #~ gids = make_nest_network(graph)


# ---------------------- #
# Pytest formatted tests #
# ---------------------- #

@pytest.mark.mpi_skip
def test_str_attr():
    ''' Check string attributes '''
    g = nngt.Graph(5)

    # set node attribute
    node_names = ["aa", "b", "c", "dddd", "eee"]

    g.new_node_attribute("name", "string", values=node_names)

    # set edges
    edges = [(0, 1), (1, 3), (1, 4), (2, 0), (3, 2), (4, 1)]

    g.new_edges(edges)

    # set edge attribute
    eattr = ["a"*i for i in range(len(edges))]

    g.new_edge_attribute("edata", "string", values=eattr)

    # check attributes
    assert list(g.node_attributes["name"]) == node_names
    assert list(g.edge_attributes["edata"]) == eattr

    # save and load string attributes
    current_dir = os.path.dirname(os.path.abspath(__file__)) + '/'

    filename = current_dir + "g.el"

    g.to_file(filename)

    h = nngt.load_from_file(filename)

    assert list(h.node_attributes["name"]) == node_names
    assert list(h.edge_attributes["edata"]) == eattr

    os.remove(filename)

    # change an attribute
    node_names[2] = "cc"
    h.set_node_attribute("name", values=node_names)

    assert not np.array_equal(h.node_attributes["name"],
                              g.node_attributes["name"])

    assert list(h.node_attributes["name"]) == node_names

    eattr[0] = "l"
    h.set_edge_attribute("edata", values=eattr)

    assert not np.array_equal(h.edge_attributes["edata"],
                              g.edge_attributes["edata"])

    assert list(h.edge_attributes["edata"]) == eattr


@pytest.mark.mpi_skip
def test_delays():
    dmin = 1.
    dmax = 8.

    d = {
        "distribution": "lin_corr", "correl_attribute": "distance",
        "lower": dmin, "upper": dmax
    }

    g = nngt.generation.distance_rule(200., nodes=100, avg_deg=10, delays=d)

    delays = g.get_delays()

    assert np.isclose(delays.min(), dmin)
    assert np.isclose(delays.max(), dmax)

    distances = g.edge_attributes["distance"]

    slope = 0.003

    g.set_delays(dmin + slope*distances)

    delays = g.get_delays()

    assert np.all(np.isclose(delays, dmin + slope*distances))


@pytest.mark.mpi_skip
def test_attributes_are_copied():
    ''' Check that the attributes returned are a copy '''
    rng = np.random.default_rng()

    nnodes = 100
    nedges = 1000

    wghts = rng.uniform(0, 5, nedges)

    g = nngt.generation.erdos_renyi(nodes=nnodes, edges=nedges, weights=wghts)

    # check weights
    ww = g.get_weights()

    assert np.all(np.isclose(wghts, ww))

    rng.shuffle(ww)

    assert np.all(np.isclose(wghts, g.get_weights()))
    assert not np.all(np.isclose(ww, g.get_weights()))

    # check edge attribute
    g.new_edge_attribute("etest", "double", values=2*ww)

    etest = g.edge_attributes["etest"]

    assert np.all(np.isclose(etest, 2*ww))

    rng.shuffle(etest)

    assert np.all(np.isclose(2*ww, g.edge_attributes["etest"]))
    assert not np.all(np.isclose(2*ww, etest))

    # check node attribute
    vv = rng.uniform(2, 3, nnodes)

    g.new_node_attribute("ntest", "double", values=vv)

    ntest = g.node_attributes["ntest"]

    assert np.all(np.isclose(ntest, vv))

    rng.shuffle(ntest)

    assert np.all(np.isclose(vv, g.node_attributes["ntest"]))
    assert not np.all(np.isclose(vv, ntest))


@pytest.mark.mpi_skip
def test_combined_attr():
    ''' Check combined attributes '''
    g = nngt.Graph(3)
    g.new_edges(((0, 1), (1, 1), (1, 2), (2, 1)), check_self_loops=False)

    ww = (0.1, 1, 0.5, 0.2)
    dd = (2., 0., 0.3, 0.3)
    rr = (0.8, 0.4, -0.5, 0.5)

    g.set_weights(ww)
    g.new_edge_attribute("distance", "double", dd)
    g.new_edge_attribute("rnd", "double", rr)

    combine = {"weight": "max", "distance": "mean", "rnd": "sum"}

    u = g.to_undirected(combine)

    assert np.all(u.get_weights() == (0.1, 1, 0.5))
    assert np.all(u.edge_attributes["distance"] == (2, 0, 0.3))
    assert np.all(u.edge_attributes["rnd"] == (0.8, 0.4, 0))


# ---------- #
# Test suite #
# ---------- #

if not nngt.get_config('mpi'):
    suite = unittest.TestLoader().loadTestsFromTestCase(TestAttributes)

    if __name__ == "__main__":
        unittest.main()
        test_str_attr()
        test_delays()
        test_attributes_are_copied()
        test_combined_attr()



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Use NNGT to analyze NEST-simulated activity of a random balanced network.
"""

import os

import numpy as np
from scipy.special import lambertw

import nngt
import nngt.generation as ng


'''
Simulation parameters
'''

num_omp = int(os.environ.get("OMP", 8))
nngt.set_config("omp", num_omp)
nngt.set_config("seeds", [10 + i for i in range(num_omp)])

dt = 0.1         # the resolution in ms
simtime = 1000.  # Simulation time in ms
delay = 1.5      # synaptic delay in ms

g = 4.0          # ratio inhibitory weight/excitatory weight
eta = 2.0        # external rate relative to threshold rate
epsilon = 0.1    # connection probability


'''
Tools
'''

def ComputePSPnorm(tauMem, CMem, tauSyn):
    a = (tauMem / tauSyn)
    b = (1.0 / tauSyn - 1.0 / tauMem)

    # time of maximum
    t_max = 1.0 / b * (-lambertw(-np.exp(-1.0 / a) / a, -1) - 1.0 / a)
    t_max = np.real(t_max)

    # maximum of PSP for current of unit amplitude
    return (np.exp(1.0) / (tauSyn * CMem * b) *
            ((np.exp(-t_max / tauMem) - np.exp(-t_max / tauSyn)) / b -
             t_max * np.exp(-t_max / tauSyn)))


'''
Network parameters
'''

order = 1000
NE = 4 * order          # number of excitatory neurons
NI = 1 * order          # number of inhibitory neurons
N_neurons = NE + NI     # number of neurons in total
N_rec = 50              # record from 50 neurons

CE = int(epsilon * NE)  # number of excitatory synapses per neuron
CI = int(epsilon * NI)  # number of inhibitory synapses per neuron
C_tot = int(CI + CE)    # total number of synapses per neuron

tauSyn = 0.5  # synaptic time constant in ms
tauMem = 20.  # time constant of membrane potential in ms
CMem = 250.   # capacitance of membrane in in pF
theta = 20.   # membrane threshold potential in mV

neuron_params = {"C_m": CMem,
                 "tau_m": tauMem,
                 "tau_syn_ex": tauSyn,
                 "tau_syn_in": tauSyn,
                 "t_ref": 2.0,
                 "E_L": 0.0,
                 "V_reset": 0.0,
                 "V_m": 0.0,
                 "V_th": theta}

J_unit = ComputePSPnorm(tauMem, CMem, tauSyn)
J = 0.8            # postsynaptic amplitude in mV
J_ex = J / J_unit  # amplitude of excitatory postsynaptic current
J_in = g * J_ex    # amplitude of inhibitory postsynaptic current

nu_th = (theta * CMem) / (J_ex * CE * np.e * tauMem * tauSyn)
nu_ex = eta * nu_th
p_rate = 400. * nu_ex * CE


'''
Create the population and network
'''

pop = nngt.NeuralPop.exc_and_inhib(
    N_neurons, en_model="iaf_psc_alpha", en_param=neuron_params,
    in_model="iaf_psc_alpha", in_param=neuron_params)

net = nngt.Network(population=pop)

ng.connect_groups(net, pop, "excitatory", "gaussian_degree", degree_type="in",
                  avg=CE, std=0.1*CE, weights=J_ex, delays=delay)

ng.connect_groups(net, pop, "inhibitory", "gaussian_degree", degree_type="in",
                  avg=CI, std=0.1*CI, weights=J_in, delays=delay)


'''
Tools to analyze the activity
'''

def coefficient_of_variation(spikes):
    from scipy.stats import variation
    times = np.array(spikes["times"])
    nrns  = np.array(spikes["senders"])

    return np.nanmean([variation(np.diff(times[nrns == i]))
                       for i in np.unique(nrns)])


def cross_correlation(spikes, time, numbins=1000):
    times = np.array(spikes["times"])
    nrns  = np.array(spikes["senders"])

    a = [np.histogram(times[nrns == x], numbins, range=[0, time])[0]
         for x in np.unique(nrns)]

    c = np.cov(a)
    std = np.sqrt(np.diag(c))
    c = np.divide(c, std[:,np.newaxis])
    c = np.divide(c, std[:,np.newaxis].T)

    return np.nanmean(c)


'''
Send the network to NEST, set noise and randomize parameters
'''

if nngt.get_config('with_nest'):
    import nest
    import nngt.simulation as ns
    from nngt.analysis import get_spikes

    nest.ResetKernel()

    nest.SetKernelStatus({"resolution": dt, "print_time": True,
                          "overwrite_files": True, 'local_num_threads': 4})

    gids = net.to_nest()

    pg = ns.set_poisson_input(gids, rate=p_rate,
                              syn_spec={"weight": J_ex, "delay": delay})

    recorders, records = ns.monitor_groups(
        ["excitatory", "inhibitory"], network=net)

    nest.Simulate(simtime)

    if nngt.get_config('with_plot'):
        ideg = net.get_degrees("in", edge_type="inhibitory")
        edeg = net.get_degrees("in", edge_type="excitatory")

        # plot the basic activity
        ns.plot_activity(
            recorders, records, network=net, show=False)
        # sort by firing rate
        ns.plot_activity(
            recorders, records, network=net, sort="firing_rate", show=False)
        # by in-degree (not working)
        ns.plot_activity(
            recorders, records, network=net, sort="in-degree", show=False)
        # k_e - k_i (good)
        ns.plot_activity(
            recorders, records, network=net, sort=edeg-ideg, show=True)

        # then show correlations
        exc_nodes = pop["excitatory"].ids
        inh_nodes = pop["inhibitory"].ids
        nngt.plot.correlation_to_attribute(
            net, (edeg-ideg)[exc_nodes], "firing_rate", nodes=exc_nodes,
            show=False)
        nngt.plot.correlation_to_attribute(
            net, (edeg-ideg)[inh_nodes], "firing_rate", nodes=inh_nodes,
            show=True)

        # print the CV and CC
        data_exc = get_spikes(recorders[0], astype="np")
        data_inh = get_spikes(recorders[1], astype="np")

        spikes = {
            "senders": list(data_exc[:, 0]) + list(data_inh[:, 0]),
            "times": list(data_exc[:, 1]) + list(data_inh[:, 1]),
        }

        print(coefficient_of_variation(spikes),
              cross_correlation(spikes, simtime))



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Simple graph generation '''

import numpy as np

import nngt
import nngt.generation as ng


# np.random.seed(0)


# ------------------- #
# Generate the graphs #
# ------------------- #

num_nodes  = 1000
avg_deg_er = 25
avg_deg_sf = 100

# random graphs
g1 = ng.erdos_renyi(nodes=num_nodes, avg_deg=avg_deg_er)

# the same graph but undirected
g2 = ng.erdos_renyi(nodes=num_nodes, avg_deg=avg_deg_er, directed=False)

# 2-step generation of a scale-free with Gaussian weight distribution
w = {
    "distribution": "gaussian",
    "avg": 60.,
    "std": 5.
}

g3 = nngt.Graph(num_nodes, weights=w)
ng.random_scale_free(2.2, 2.9, avg_deg=avg_deg_sf, from_graph=g3)

# same in 1 step
g4 = ng.random_scale_free(
    2.2, 2.9, avg_deg=avg_deg_sf, nodes=num_nodes, weights=w)


# ----------------- #
# Check the results #
# ----------------- #

assert np.isclose(avg_deg_er, np.average(g1.get_degrees('in')), 1e-4)
assert np.isclose(avg_deg_sf, np.average(g3.get_degrees('in')), 1e-4)
assert np.isclose(avg_deg_sf, np.average(g4.get_degrees('in')), 1e-4)

print(
    "Erdos-Renyi: requested average degree of {}; got {} for directed graph "
    "and {} for undirected one.".format(
        avg_deg_er, g1.edge_nb() / float(num_nodes),
        g2.edge_nb() / float(num_nodes))
)

if nngt.get_config('with_plot'):
    from nngt.plot import degree_distribution

    degree_distribution(
        g4, deg_type=["in", "out"], logx=True, logy=True, show=True)




#-*- coding:utf-8 -*-
#
# generation/graph_connectivity.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Connectivity generators for nngt.Graph """

from copy import deepcopy
import logging

import numpy as np

import nngt
from nngt.geometry.geom_utils import conversion_magnitude
from nngt.lib.connect_tools import _set_options
from nngt.lib.logger import _log_message
from nngt.lib.test_functions import (mpi_checker, mpi_random, deprecated,
                                     on_master_process)


__all__ = [
    'all_to_all',
    'circular',
    'distance_rule',
    'erdos_renyi',
    'fixed_degree',
    'from_degree_list',
    'gaussian_degree',
    'newman_watts',
    'random_scale_free',
    'price_scale_free',
    'watts_strogatz',
]


# do default import

from .connect_algorithms import *

# try to import multithreaded or mpi algorithms

using_mt_algorithms = False

if nngt.get_config("multithreading"):
    logger = logging.getLogger(__name__)
    try:
        from .cconnect import *
        using_mt_algorithms = True
        _log_message(logger, "DEBUG",
                     "Using multithreaded algorithms compiled on install.")
        nngt.set_config('multithreading', True, silent=True)
    except Exception as e:
        try:
            import cython
            import pyximport

            try:
                from mpi4py import MPI
                comm = MPI.COMM_WORLD
                comm.bcast(pyximport.install(language_level=3))
            except:
                if nngt.get_config("mpi"):
                    raise RuntimeError("Cannot safely compile with MPI.")

                pyximport.install(language_level=3)

            # wait for compilation to finish
            nngt.lib.mpi_barrier()

            if on_master_process():
                from .cconnect import *
    
            nngt.lib.mpi_barrier()

            from .cconnect import *

            using_mt_algorithms = True

            _log_message(logger, "DEBUG", str(e) + "\n\tCompiled "
                         "multithreaded algorithms on-the-run.")

            nngt.set_config('multithreading', True, silent=True)
        except Exception as e2:
            _log_message(
                logger, "WARNING", str(e) + "\n\t" + str(e2) + "\n\t"
                "Cython import failed, using non-multithreaded algorithms.")
            nngt._config['multithreading'] = False

if nngt.get_config("mpi"):
    try:
        from .mpi_connect import *
        nngt._config['mpi'] = True
    except ImportError as e:
        nngt._config['mpi'] = False
        raise e


# ----------------------------- #
# Specific degree distributions #
# ----------------------------- #

def all_to_all(nodes=0, weighted=True, directed=True, multigraph=False,
               name="AllToAll", shape=None, positions=None, population=None,
               **kwargs):
    """
    Generate a graph where all nodes are connected.

    .. versionadded:: 1.0

    Parameters
    ----------
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    reciprocity : double, optional (default: -1 to let it free)
        Fraction of edges that are bidirectional  (only for directed graphs
        -- undirected graphs have a reciprocity of  1 by definition)
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).

    Note
    ----
    `nodes` is required unless `population` is provided.

    Returns
    -------
    graph_all : :class:`~nngt.Graph`, or subclass
        A new generated graph.
    """
    nodes = nodes if population is None else population.size

    graph_all = nngt.Graph(name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_all, population, shape, positions)

    # add edges
    if nodes > 1:
        ids = np.arange(nodes, dtype=np.uint)
        edges = _all_to_all(ids, ids, directed=directed, multigraph=multigraph)
        graph_all.new_edges(edges, check_duplicates=False,
                            check_self_loops=False, check_existing=False)

    graph_all._graph_type = "all_to_all"

    return graph_all


@mpi_random
def from_degree_list(degrees, degree_type='in', weighted=True,
                     directed=True, multigraph=False, name="DL",
                     shape=None, positions=None, population=None,
                     from_graph=None, **kwargs):
    """
    Generate a random graph from a given list of degrees.

    Parameters
    ----------
    degrees : list
        The list of degrees for each node in the graph.
    degree_type : str, optional (default: 'in')
        The type of the fixed degree, among ``'in'``, ``'out'`` or ``'total'``.
        @todo `'total'` not implemented yet.
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        @todo: only for directed graphs for now. Whether the graph is directed
        or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_dl : :class:`~nngt.Graph`, or subclass
        A new generated graph or the modified `from_graph`.
    """
    # set node number and library graph
    graph_dl = from_graph
    nodes    = len(degrees)

    if "nodes" in kwargs:
        assert kwargs["nodes"] == nodes, \
            "Invalid `nodes` entry: the number of nodes should " \
            "be ``len(degrees)``."
        del kwargs["nodes"]

    if graph_dl is not None:
        nodes = graph_dl.node_nb()
        graph_dl.clear_all_edges()
    else:
        nodes = population.size if population is not None else nodes
        graph_dl = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_dl, population, shape, positions)

    # add edges
    ia_edges = None

    if nodes > 1:
        ids = np.arange(nodes, dtype=np.uint)
        ia_edges = _from_degree_list(ids, ids, degrees, degree_type,
                                     directed=directed, multigraph=multigraph)
        # check for None if MPI
        if ia_edges is not None:
            graph_dl.new_edges(ia_edges, check_duplicates=False,
                               check_self_loops=False, check_existing=False)

    graph_dl._graph_type = "from_{}_degree_list".format(degree_type)

    return graph_dl


@mpi_random
def fixed_degree(degree, degree_type='in', nodes=0, reciprocity=-1.,
                 weighted=True, directed=True, multigraph=False, name="FD",
                 shape=None, positions=None, population=None, from_graph=None,
                 **kwargs):
    """
    Generate a random graph with constant in- or out-degree.

    Parameters
    ----------
    degree : int
        The value of the constant degree.
    degree_type : str, optional (default: 'in')
        The type of the fixed degree, among ``'in'``, ``'out'`` or ``'total'``.

        @todo
            `'total'` not implemented yet.

    nodes : int, optional (default: None)
        The number of nodes in the graph.
    reciprocity : double, optional (default: -1 to let it free)
        @todo: not implemented yet. Fraction of edges that are bidirectional
        (only for directed graphs -- undirected graphs have a reciprocity of
        1 by definition)
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        @todo: only for directed graphs for now. Whether the graph is directed
        or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Note
    ----
    `nodes` is required unless `from_graph` or `population` is provided.
    If an `from_graph` is provided, all preexistant edges in the
    object will be deleted before the new connectivity is implemented.

    Returns
    -------
    graph_fd : :class:`~nngt.Graph`, or subclass
        A new generated graph or the modified `from_graph`.
    """
    # set node number and library graph
    graph_fd = from_graph

    if graph_fd is not None:
        nodes = graph_fd.node_nb()
        graph_fd.clear_all_edges()
    else:
        nodes = population.size if population is not None else nodes
        graph_fd = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_fd, population, shape, positions)

    # add edges
    ia_edges = None

    if nodes > 1:
        ids = np.arange(nodes, dtype=np.uint)
        ia_edges = _fixed_degree(
            ids, ids, degree, degree_type, reciprocity=reciprocity,
            directed=directed, multigraph=multigraph)
        # check for None if MPI
        if ia_edges is not None:
            graph_fd.new_edges(ia_edges, check_duplicates=False,
                               check_self_loops=False, check_existing=False)

    graph_fd._graph_type = "fixed_{}_degree".format(degree_type)

    return graph_fd


@mpi_random
def gaussian_degree(avg, std, degree_type='in', nodes=0, reciprocity=-1.,
                    weighted=True, directed=True, multigraph=False, name="GD",
                    shape=None, positions=None, population=None,
                    from_graph=None, **kwargs):
    """
    Generate a random graph with constant in- or out-degree.

    Parameters
    ----------
    avg : float
        The value of the average degree.
    std : float
        The standard deviation of the Gaussian distribution.
    degree_type : str, optional (default: 'in')
        The type of the fixed degree, among 'in', 'out' or 'total' (or the
        full version: 'in-degree'...)
        @todo: Implement 'total' degree
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    reciprocity : double, optional (default: -1 to let it free)
        @todo: not implemented yet. Fraction of edges that are bidirectional
        (only for directed graphs -- undirected graphs have a reciprocity of
        1 by definition)
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        @todo: only for directed graphs for now. Whether the graph is directed
        or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_gd : :class:`~nngt.Graph`, or subclass
        A new generated graph or the modified `from_graph`.

    Note
    ----
    `nodes` is required unless `from_graph` or `population` is provided.
    If an `from_graph` is provided, all preexistant edges in the object
    will be deleted before the new connectivity is implemented.
    """
    # set node number and library graph
    graph_gd = from_graph

    if graph_gd is not None:
        nodes = graph_gd.node_nb()
        graph_gd.clear_all_edges()
    else:
        nodes    = population.size if population is not None else nodes
        graph_gd = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_gd, population, shape, positions)

    # add edges
    ia_edges = None
    if nodes > 1:
        ids = np.arange(nodes, dtype=np.uint)
        ia_edges = _gaussian_degree(
            ids, ids, avg, std, degree_type, reciprocity=reciprocity,
            directed=directed, multigraph=multigraph)
        # check for None if MPI
        if ia_edges is not None:
            graph_gd.new_edges(ia_edges, check_duplicates=False,
                               check_self_loops=False, check_existing=False)

    graph_gd._graph_type = "gaussian_{}_degree".format(degree_type)

    return graph_gd


# ----------- #
# Erdos-Renyi #
# ----------- #

def erdos_renyi(density=None, nodes=0, edges=None, avg_deg=None,
                reciprocity=-1., weighted=True, directed=True,
                multigraph=False, name="ER", shape=None, positions=None,
                population=None, from_graph=None, **kwargs):
    """
    Generate a random graph as defined by Erdos and Renyi but with a
    reciprocity that can be chosen.

    Parameters
    ----------
    density : double, optional (default: -1.)
        Structural density given by `edges / nodes`:math:`^2`. It is also the
        probability for each possible edge in the graph to exist.
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    edges : int (optional)
        The number of edges between the nodes
    avg_deg : double, optional
        Average degree of the neurons given by `edges / nodes`.
    reciprocity : double, optional (default: -1 to let it free)
        Fraction of edges that are bidirectional (only for
        directed graphs -- undirected graphs have a reciprocity of 1 by
        definition)
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_er : :class:`~nngt.Graph`, or subclass
        A new generated graph or the modified `from_graph`.

    Note
    ----
    `nodes` is required unless `from_graph` or `population` is provided.
    If an `from_graph` is provided, all preexistant edges in the
    object will be deleted before the new connectivity is implemented.
    """
    # set node number and library graph
    graph_er = from_graph

    if graph_er is not None:
        nodes = graph_er.node_nb()
        graph_er.clear_all_edges()
    else:
        nodes = population.size if population is not None else nodes
        graph_er = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_er, population, shape, positions)

    # add edges
    ia_edges = None

    if nodes > 1:
        ids = range(nodes)
        ia_edges = _erdos_renyi(ids, ids, density, edges, avg_deg, reciprocity,
                                directed, multigraph)
        graph_er.new_edges(ia_edges, check_duplicates=False,
                           check_self_loops=False, check_existing=False)

    graph_er._graph_type = "erdos_renyi"

    return graph_er


# ----------------- #
# Scale-free models #
# ----------------- #

def random_scale_free(in_exp, out_exp, nodes=0, density=None, edges=None,
                      avg_deg=None, reciprocity=0., weighted=True,
                      directed=True, multigraph=False, name="RandomSF",
                      shape=None, positions=None, population=None,
                      from_graph=None, **kwargs):
    """
    Generate a free-scale graph of given reciprocity and otherwise
    devoid of correlations.

    Parameters
    ----------
    in_exp : float
        Absolute value of the in-degree exponent :math:`\gamma_i`, such that
        :math:`p(k_i) \propto k_i^{-\gamma_i}`
    out_exp : float
        Absolute value of the out-degree exponent :math:`\gamma_o`, such that
        :math:`p(k_o) \propto k_o^{-\gamma_o}`
    nodes : int, optional (default: 0)
        The number of nodes in the graph.
    density: double, optional
        Structural density given by `edges / (nodes*nodes)`.
    edges : int optional
        The number of edges between the nodes
    avg_deg : double, optional
        Average degree of the neurons given by `edges / nodes`.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes. can contain multiple edges between two
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`)
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_fs : :class:`~nngt.Graph`

    Note
    ----
    As reciprocity increases, requested values of `in_exp` and `out_exp`
    will be less and less respected as the distribution will converge to a
    common exponent :math:`\gamma = (\gamma_i + \gamma_o) / 2`.
    Parameter `nodes` is required unless `from_graph` or `population` is
    provided.
    """
    # set node number and library graph
    graph_rsf = from_graph
    if graph_rsf is not None:
        nodes = graph_rsf.node_nb()
        graph_rsf.clear_all_edges()
    else:
        nodes = population.size if population is not None else nodes
        graph_rsf = nngt.Graph(
            name=name,nodes=nodes,directed=directed,**kwargs)

    _set_options(graph_rsf, population, shape, positions)

    # add edges
    if nodes > 1:
        ids = range(nodes)
        ia_edges = _random_scale_free(ids, ids, in_exp, out_exp, density,
                          edges, avg_deg, reciprocity, directed, multigraph)
        graph_rsf.new_edges(ia_edges, check_duplicates=False,
                            check_self_loops=False, check_existing=False)

    graph_rsf._graph_type = "random_scale_free"

    return graph_rsf


def price_scale_free(m, c=None, gamma=1, nodes=0, reciprocity=0, weighted=True,
                     directed=True, multigraph=False, name="PriceSF",
                     shape=None, positions=None, population=None, **kwargs):
    r'''
    Generate a Price graph model (Barabasi-Albert if undirected).

    Parameters
    ----------
    m : int
        The number of edges each new node will make.
    c : double, optional (0 if undirected, else 1)
        Constant added to the probability of a vertex receiving an edge.
    gamma : double, optional (default: 1)
        Preferential attachment power.
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    reciprocity : float, optional (default: 0)
        Reciprocity of the graph (between 0 and 1). For directed graphs, this
        will be the probability of the target node connecting back to the
        source node when a new edge is added.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).

    Returns
    -------
    graph_price : :class:`~nngt.Graph` or subclass.

    Note
    ----
    `nodes` is required unless `population` is provided.

    Notes
    -----
    The (generalized) Price network is either a directed or undirected graph
    (the latter is better known as the Barabási-Albert network).
    It is generated via a growth process, adding a new node at each step and
    connecting it to :math:`m` previous nodes, chosen with probability:

    .. math::

        p \propto k^\gamma + c

    where :math:`k` is the (in-)degree of the vertex.

    We must therefore have :math:`c \ge 0` for directed graphs and
    :math:`c > -1` for undirected graphs.

    If the `reciprocity` :math:`r` is non-zero, each targeted node reciprocates
    the connection with probability :math:`r`.
    Expected reciprocity of the final graph is :math:`2r / (1 + r)`.

    If :math:`\gamma=1`, and `reciprocity` is zero, the tail of resulting
    in-degree distribution of the directed case is given by

    .. math::

        P_{k_{in}} \sim k_{in}^{-(2 + c/m)},

    or for the undirected case

    .. math::

        P_{k} \sim k^{-(3 + c/m)}.

    However, if :math:`\gamma \ne 1`, the in-degree distribution is not
    scale-free.
    '''
    c = c if c is not None else 1 if directed else 0

    # set node number and library graph
    nodes = population.size if population is not None else nodes

    graph_psf = nngt.Graph(
        name=name, nodes=nodes, directed=directed, weighted=weighted,
        **kwargs)

    _set_options(graph_psf, population, shape, positions)

    # add edges
    if nodes > 1:
        ids = range(nodes)
        edges = _price_scale_free(ids, m, c, gamma, reciprocity, directed,
                                     multigraph)

        graph_psf.new_edges(edges, check_duplicates=False,
                            check_self_loops=False, check_existing=False)

    graph_psf._graph_type = "price_scale_free"

    return graph_psf


# -------------- #
# Circular graph #
# -------------- #

def circular(coord_nb, reciprocity=1., reciprocity_choice="random", nodes=0,
             weighted=True, directed=True, multigraph=False, name="Circular",
             shape=None, positions=None, population=None, from_graph=None,
             **kwargs):
    '''
    Generate a circular graph.

    The nodes are placed on a circle and connected to their `coord_nb` closest
    neighbours.
    If the graph is directed, the number of connections depends on the value
    of `reciprocity`: if ``reciprocity == 0.``, then only half of all possible
    connections will be created, so that no bidirectional edges exist; on the
    other hand, for ``reciprocity == 1.``, all possible edges are created; for
    intermediate values of `reciprocity`, the number of edges increases
    linearly as ``0.5*(1 + reciprocity / (2 - reciprocity))*nodes*coord_nb``.

    Parameters
    ----------
    coord_nb : int
        The number of neighbours for each node on the initial topological
        lattice (must be even).
    reciprocity : double, optional (default: 1.)
        Proportion of reciprocal edges in the graph.
    reciprocity_choice : str, optional (default: "random")
        How reciprocal edges should be chosen, which can be either "random" or
        "closest". If the latter option is used, then connections
        between first neighbours are rendered reciprocal first, then between
        second neighbours, etc.
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    density: double, optional (default: 0.1)
        Structural density given by `edges` / (`nodes`*`nodes`).
    edges : int (optional)
        The number of edges between the nodes
    avg_deg : double, optional
        Average degree of the neurons given by `edges` / `nodes`.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_circ : :class:`~nngt.Graph` or subclass
    '''
    if multigraph:
        raise ValueError("`multigraph` is not supported for circular graphs.")

    # set node number and library graph
    graph_circ = from_graph

    if graph_circ is not None:
        nodes = graph_circ.node_nb()
    else:
        nodes = population.size if population is not None else nodes
        graph_circ = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_circ, population, shape, positions)

    # add edges
    if nodes > 1:
        ids   = range(nodes)
        edges = _circular(ids, ids, coord_nb, reciprocity, directed,
                          reciprocity_choice=reciprocity_choice)

        graph_circ.new_edges(edges, check_duplicates=False,
                             check_self_loops=False, check_existing=False)

    graph_circ._graph_type = "circular"

    return graph_circ
    

# ------------------ #
# Small-world models #
# ------------------ #

def newman_watts(coord_nb, proba_shortcut=None, reciprocity_circular=1.,
                 reciprocity_choice_circular="random", nodes=0, edges=None,
                 weighted=True, directed=True, multigraph=False, name="NW",
                 shape=None, positions=None, population=None, from_graph=None,
                 **kwargs):
    """
    Generate a (potentially small-world) graph using the Newman-Watts
    algorithm.

    For directed networks, the reciprocity of the initial circular network can
    be chosen.

    .. versionchanged:: 2.0
        Added the `reciprocity_circular` and `reciprocity_choice_circular`
        options.

    Parameters
    ----------
    coord_nb : int
        The number of neighbours for each node on the initial topological
        lattice (must be even).
    proba_shortcut : double, optional
        Probability of adding a new random (shortcut) edge for each existing
        edge on the initial lattice.
        If `edges` is provided, then will be computed automatically as
        ``edges / (coord_nb * nodes * (1 + reciprocity_circular) / 2)``
    reciprocity_circular : double, optional (default: 1.)
        Proportion of reciprocal edges in the initial circular graph.
    reciprocity_choice_circular : str, optional (default: "random")
        How reciprocal edges should be chosen in the initial circular graph.
        This can be either "random" or "closest". If the latter option
        is used, then connections between first neighbours are rendered
        reciprocal first, then between second neighbours, etc.
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    edges : int (optional)
        The number of edges between the nodes.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_nw : :class:`~nngt.Graph` or subclass

    Note
    ----
    `nodes` is required unless `from_graph` or `population` is provided.
    """
    if multigraph:
        raise ValueError("`multigraph` is not supported for Watts-Strogatz.")

    # set node number and library graph
    graph_nw = from_graph

    if graph_nw is not None:
        nodes = graph_nw.node_nb()
    else:
        nodes = population.size if population is not None else nodes
        graph_nw = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_nw, population, shape, positions)

    # add edges
    if nodes > 1:
        ids = range(nodes)

        ia_edges = _newman_watts(
            ids, ids, coord_nb, proba_shortcut, reciprocity_circular,
            reciprocity_choice_circular=reciprocity_choice_circular,
            edges=edges, directed=directed)

        graph_nw.new_edges(ia_edges, check_duplicates=False,
                           check_self_loops=False, check_existing=False)

    graph_nw._graph_type = "watts_strogatz"

    return graph_nw


def watts_strogatz(coord_nb, proba_shortcut=None, reciprocity_circular=1.,
                   reciprocity_choice_circular="random", shuffle="random",
                   nodes=0, weighted=True, directed=True, multigraph=False,
                   name="WS", shape=None, positions=None, population=None,
                   from_graph=None,  **kwargs):
    """
    Generate a (potentially small-world) graph using the Watts-Strogatz
    algorithm.

    For directed networks, the reciprocity of the initial circular network can
    be chosen.

    .. versionadded:: 2.0

    Parameters
    ----------
    coord_nb : int
        The number of neighbours for each node on the initial topological
        lattice (must be even).
    proba_shortcut : double, optional
        Probability of adding a new random (shortcut) edge for each existing
        edge on the initial lattice.
        If `edges` is provided, then will be computed automatically as
        ``edges / (coord_nb * nodes * (1 + reciprocity_circular) / 2)``
    reciprocity_circular : double, optional (default: 1.)
        Proportion of reciprocal edges in the initial circular graph.
    reciprocity_choice_circular : str, optional (default: "random")
        How reciprocal edges should be chosen in the initial circular graph.
        This can be either "random" or "closest". If the latter option
        is used, then connections between first neighbours are rendered
        reciprocal first, then between second neighbours, etc.
    shuffle : str, optional (default: 'random')
        Whether to shuffle only 'targets' (out-degree of all nodes remains
        constant), 'sources' (in-degree remains constant), or randomly the
        source or the target for each edge ('random') in the case of directed
        graphs.
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "ER")
        Name of the created graph.
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D or 3D array containing the positions of the neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.

    Returns
    -------
    graph_nw : :class:`~nngt.Graph` or subclass

    Note
    ----
    `nodes` is required unless `from_graph` or `population` is provided.
    """
    if multigraph:
        raise ValueError("`multigraph` is not supported for Newman-Watts.")

    # set node number and library graph
    graph_nw = from_graph
    if graph_nw is not None:
        nodes = graph_nw.node_nb()
    else:
        nodes = population.size if population is not None else nodes
        graph_nw = nngt.Graph(
            name=name, nodes=nodes, directed=directed, **kwargs)

    _set_options(graph_nw, population, shape, positions)

    # add edges
    if nodes > 1:
        ids = range(nodes)

        ia_edges = _watts_strogatz(
            ids, ids, coord_nb, proba_shortcut, reciprocity_circular,
            reciprocity_choice_circular, shuffle, directed=directed)

        graph_nw.new_edges(ia_edges, check_duplicates=False,
                           check_self_loops=False, check_existing=False)

    graph_nw._graph_type = "newman_watts"

    return graph_nw


# --------------------- #
# Distance-based models #
# --------------------- #

@mpi_random
def distance_rule(scale, rule="exp", shape=None, neuron_density=1000.,
                  max_proba=-1., nodes=0, density=None, edges=None,
                  avg_deg=None, unit='um', weighted=True, directed=True,
                  multigraph=False, name="DR", positions=None, population=None,
                  from_graph=None, **kwargs):
    """
    Create a graph using a 2D distance rule to create the connection between
    neurons. Available rules are linear and exponential.

    Parameters
    ----------
    scale : float
        Characteristic scale for the distance rule. E.g for linear distance-
        rule, :math:`P(i,j) \propto (1-d_{ij}/scale))`, whereas for the
        exponential distance-rule, :math:`P(i,j) \propto e^{-d_{ij}/scale}`.
    rule : string, optional (default: 'exp')
        Rule that will be apply to draw the connections between neurons.
        Choose among "exp" (exponential), "gaussian" (Gaussian), or
        "lin" (linear).
    shape : :class:`~nngt.geometry.Shape`, optional (default: None)
        Shape of the neurons' environment. If not specified, a square will be
        created with the appropriate dimensions for the number of neurons and
        the neuron spatial density.
    neuron_density : float, optional (default: 1000.)
        Density of neurons in space (:math:`neurons \cdot mm^{-2}`).
    nodes : int, optional (default: None)
        The number of nodes in the graph.
    p : float, optional
        Normalization factor for the distance rule; it is equal to the
        probability of connection when testing a node at zero distance.
    density: double, optional
        Structural density given by `edges` / (`nodes` * `nodes`).
    edges : int, optional
        The number of edges between the nodes
    avg_deg : double, optional
        Average degree of the neurons given by `edges` / `nodes`.
    unit : string (default: 'um')
        Unit for the length `scale` among 'um' (:math:`\mu m`), 'mm', 'cm',
        'dm', 'm'.
    weighted : bool, optional (default: True)
        Whether the graph edges have weights.
    directed : bool, optional (default: True)
        Whether the graph is directed or not.
    multigraph : bool, optional (default: False)
        Whether the graph can contain multiple edges between two
        nodes.
    name : string, optional (default: "DR")
        Name of the created graph.
    positions : :class:`numpy.ndarray`, optional (default: None)
        A 2D (N, 2) or 3D (N, 3) shaped array containing the positions of the
        neurons in space.
    population : :class:`~nngt.NeuralPop`, optional (default: None)
        Population of neurons defining their biological properties (to create a
        :class:`~nngt.Network`).
    from_graph : :class:`Graph` or subclass, optional (default: None)
        Initial graph whose nodes are to be connected.
    """
    distance = []
    # convert neuronal density in (mu m)^2
    neuron_density *= conversion_magnitude(unit, 'mm')**2
    # set node number and library graph
    graph_dr = from_graph
    if graph_dr is not None:
        nodes = graph_dr.node_nb()
        graph_dr.clear_all_edges()
    else:
        nodes = population.size if population is not None else nodes
    # check shape
    if shape is None:
        h = w = np.sqrt(float(nodes) / neuron_density)
        shape = nngt.geometry.Shape.rectangle(h, w)
    if graph_dr is None:
        graph_dr = nngt.SpatialGraph(
            name=name, nodes=nodes, directed=directed, shape=shape,
            positions=positions, **kwargs)
    else:
        Graph.make_spatial(graph_dr, shape, positions=positions)
    positions = np.array(graph_dr.get_positions().T, dtype=np.float32)
    # set options (graph has already been made spatial)
    _set_options(graph_dr, population, None, None)
    # add edges
    ia_edges = None
    conversion_factor = conversion_magnitude(shape.unit, unit)
    if unit != shape.unit:
        positions = np.multiply(conversion_factor, positions, dtype=np.float32)
    if nodes > 1:
        ids = np.arange(0, nodes, dtype=np.uint)
        ia_edges = _distance_rule(
            ids, ids, density, edges, avg_deg, scale, rule, max_proba, shape,
            positions, directed, multigraph, distance=distance, **kwargs)
        attr = {'distance': distance}
        # check for None if MPI
        if ia_edges is not None:
            graph_dr.new_edges(ia_edges, attributes=attr,
                               check_duplicates=False, check_self_loops=False,
                               check_existing=False)

    graph_dr._graph_type = "{}_distance_rule".format(rule)
    return graph_dr


# -------------------- #
# Polyvalent generator #
# -------------------- #

_di_generator = {
    "all_to_all": all_to_all,
    "circular": circular,
    "distance_rule": distance_rule,
    "erdos_renyi": erdos_renyi,
    "fixed_degree": fixed_degree,
    "from_degree_list": from_degree_list,
    "gaussian_degree": gaussian_degree,
    "newman_watts": newman_watts,
    "price_scale_free": price_scale_free,
    "random_scale_free": random_scale_free,
    "watts_strogatz": watts_strogatz,
}


def generate(di_instructions, **kwargs):
    '''
    Generate a :class:`~nngt.Graph` or one of its subclasses from a ``dict``
    containing all the relevant informations.

    Parameters
    ----------
    di_instructions : ``dict``
        Dictionary containing the instructions to generate the graph. It must
        have at least ``"graph_type"`` in its keys, with a value among
        ``"distance_rule", "erdos_renyi", "fixed_degree", "newman_watts",
        "price_scale_free", "random_scale_free"``. Depending on the type,
        `di_instructions` should also contain at least all non-optional
        arguments of the generator function.

    See also
    --------
    :mod:`~nngt.generation`
    '''
    graph_type = di_instructions["graph_type"]
    instructions = deepcopy(di_instructions)

    del instructions["graph_type"]
    
    instructions.update(kwargs)

    return _di_generator[graph_type](**instructions)



#-*- coding:utf-8 -*-
#
# simulation/nest_utils.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# nest_utils.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

""" Utility functions to monitor NEST simulated activity """

try:
    from io import BytesIO
except ImportError:
    from BytesIO import BytesIO

import nest
import numpy as np

from nngt.lib import InvalidArgument, nonstring_container
from nngt.lib.rng_tools import _generate_random
from nngt.lib.sorting import _sort_groups


try:
    from nest import NodeCollection
    nest_version = 3
    spike_rec = 'spike_recorder'
except ImportError:
    nest_version = 2
    spike_rec = 'spike_detector'


__all__ = [
    'monitor_groups',
    'monitor_nodes',
    'randomize_neural_states',
    'save_spikes',
    'set_minis',
    'set_noise',
    'set_poisson_input',
    'set_step_currents',
]


#-----------------------------------------------------------------------------#
# Inducing activity
#------------------------
#

def set_noise(gids, mean, std):
    '''
    Submit neurons to a current white noise.

    Parameters
    ----------
    gids : tuple
        NEST gids of the target neurons.
    mean : float
        Mean current value.
    std : float
        Standard deviation of the current

    Returns
    -------
    noise : tuple
        The NEST gid of the noise_generator.
    '''
    noise = nest.Create("noise_generator", params={"mean": mean, "std": std},
                        _warn=False)

    nest.Connect(noise, _get_nest_gids(gids), _warn=False)

    return noise


def set_poisson_input(gids, rate, syn_spec=None, **kwargs):
    '''
    Submit neurons to a Poissonian rate of spikes.

    .. versionchanged:: 2.0
        Added `kwargs`.

    Parameters
    ----------
    gids : tuple
        NEST gids of the target neurons.
    rate : float
        Rate of the spike train (in Hz).
    syn_spec : dict, optional (default: static synapse with weight 1)
        Properties of the connection between the ``poisson_generator`` object
        and the target neurons.
    **kwargs : dict
        Other optional parameters for the `poisson_generator`.

    Returns
    -------
    poisson_input : tuple
        The NEST gid of the ``poisson_generator``.
    '''
    params = {"rate": rate}
    params.update(kwargs)

    poisson_input = nest.Create(
        "poisson_generator", params=params, _warn=False)

    gids = _get_nest_gids(gids)

    nest.Connect(poisson_input, gids, syn_spec=syn_spec, _warn=False)

    return poisson_input


def set_minis(network, base_rate, weight, syn_type=1, nodes=None, gids=None):
    '''
    Mimick spontaneous release of neurotransmitters, called miniature PSCs or
    "minis" that can occur at excitatory (mEPSCs) or inhibitory (mIPSCs)
    synapses.
    These minis consists in only a fraction of the usual strength of a spike-
    triggered PSC and can be modeled by a Poisson process.
    This Poisson process occurs independently at every synapse of a neuron, so
    a neuron receiving :math:`k` inputs will be subjected to these events with
    a rate :math:`k*\\lambda`, where :math:`\\lambda` is the base rate.

    Parameters
    ----------
    network : :class:`~nngt.Network` object
        Network on which the minis should be simulated.
    base_rate : float
        Rate for the Poisson process on one synapse (:math:`\\lambda`), in Hz.
    weight : float or array of size N
        Amplitude of a minitature post-synaptic event.
    syn_type : int, optional (default: 1)
        Synaptic type of the noisy connections. By default, mEPSCs are
        generated, by taking into account only the excitatory degrees and
        synaptic weights. To setup mIPSCs, used `syn_type=-1`.
    nodes : array-like (size N), optional (default: all nodes)
        NNGT ids of the neurons that should be subjected to minis.
    gids : array-like (size N), optional (default: all neurons)
        NEST gids of the neurons that should be subjected to minis.

    Note
    ----
    `nodes` and `gids` are not compatible, only one one the two arguments can
    be used in any given call to `set_minis`.
    '''
    assert network.nest_gids is not None, \
        "Create the NEST network before calling `set_minis`."

    # find the target nodes' gids
    if gids is not None and nodes is not None:
        raise InvalidArgument('Only one of `nodes` and `gids` can be set.')
    elif nodes is None and gids is None:
        nodes = range(0, network.node_nb())
        gids = network.nest_gids
    elif gids is not None:
        nodes = [network.id_from_nest_gid(gid) for gid in gids]
    elif nodes is not None:
        gids = network.nest_gids[nodes]

    # convert gids if necessary
    gids = _get_nest_gids(gids)

    if nonstring_container(weight):
        assert len(weight) == len(gids)
    else:
        weight = [weight for _ in range(len(gids))]

    # get the unique degrees and create one poisson_generator per degree
    degrees    = network.get_degrees("in", edge_type=syn_type)
    deg_set    = set(degrees)
    map_deg_pg = {d: i for i, d in enumerate(deg_set)}
    pgs        = nest.Create("poisson_generator", len(deg_set), _warn=False)

    for d, pg in zip(deg_set, pgs):
        if nest_version == 2:
            pg = (pg,)

        nest.SetStatus(pg, {"rate": d*base_rate}, _warn=False)

    # connect
    for i, n in enumerate(nodes):
        gid, d = gids[i], degrees[n]

        pg = pgs[map_deg_pg[d]]

        if nest_version == 2:
            gid = (gid,)
            pg = (pg,)

        w  = weight[i]

        nest.Connect(pg, gid, syn_spec={'weight': w}, _warn=False)


def set_step_currents(gids, times, currents):
    '''
    Set step-current excitations

    Parameters
    ----------
    gids : tuple
        NEST gids of the target neurons.
    times : list or :class:`numpy.ndarray`
        List of the times where the current will change (by default the current
        generator is initiated at I=0. for t=0.)
    currents : list or :class:`numpy.ndarray`
        List of the new current value after the associated time value in
        `times`.

    Returns
    -------
    noise : tuple
        The NEST gid of the noise_generator.
    '''
    if len(times) != len(currents):
        raise InvalidArgument('Length of `times` and `currents` must be the '
                              'same')

    params = {"amplitude_times": times, "amplitude_values":currents}

    scg = nest.Create("step_current_generator", params=params, _warn=False)

    gids = _get_nest_gids(gids)

    nest.Connect(scg, gids, _warn=False)

    return scg


def randomize_neural_states(network, instructions, groups=None, nodes=None,
                            make_nest=False):
    '''
    Randomize the neural states according to the instructions.

    .. versionchanged:: 0.8
        Changed `ids` to `nodes` argument.

    Parameters
    ----------
    network : :class:`~nngt.Network` subclass instance
        Network that will be simulated.
    instructions : dict
        Variables to initialize. Allowed keys are "V_m" and "w". Values are
        3-tuples of type ``("distrib_name", double, double)``.
    groups : list of :class:`~nngt.NeuralGroup`, optional (default: None)
        If provided, only the neurons belonging to these groups will have their
        properties randomized.
    nodes : array-like, optional (default: all neurons)
        NNGT ids of the neurons that will have their status randomized.
    make_nest : bool, optional (default: False)
        If ``True`` and network has not been converted to NEST, automatically
        generate the network, else raises an exception.

    Example
    -------

    .. code-block:: python

        instructions = {
            "V_m": ("uniform", -80., -60.),
            "w": ("normal", 50., 5.)
        }
    '''
    # check whether network is in NEST
    if network._nest_gids is None:
        if make_nest:
            network.to_nest()
        else:
            raise AttributeError(
                '`network` has not been sent to NEST yet.')

    gids = []

    if nodes is not None and groups is not None:
        raise InvalidArgument('`nodes` and `groups` cannot be set together.')
    elif groups is not None:
        for group in groups:
            gids.extend(group.nest_gids)
        gids = list(set(gids))
    else:
        gids.extend(
            network.nest_gids if nodes is None else network.nest_gids[nodes])

    # convert gids
    gids = _get_nest_gids(gids)

    num_neurons = len(gids)

    for key, val in instructions.items():
        state = _generate_random(num_neurons, val)

        # set the values in NEST
        nest.SetStatus(gids, key, state, _warn=False)

        if nodes is None:
            nodes = network.id_from_nest_gid(np.asarray(gids))

        # store the values in the node attributes
        if key not in ("V_m", "w"):
            if key not in network.node_attributes:
                network.new_node_attribute(key, "double", val=np.NaN)
            network.set_node_attribute(
                key, values=state, nodes=nodes, value_type="double")


# ----------------------- #
# Monitoring the activity #
# ----------------------- #

def monitor_groups(group_names, network, nest_recorder=None, params=None):
    '''
    Monitoring the activity of nodes in the network.

    Parameters
    ----------
    group_name : list of strings
        Names of the groups that should be recorded.
    network : :class:`~nngt.Network` or subclass
        Network which population will be used to differentiate groups.
    nest_recorder : strings or list, optional (default: spike recorder)
        Device(s) to monitor the network.
    params : dict or list of, optional (default: `{}`)
        Dictionarie(s) containing the parameters for each recorder (see
        `NEST documentation <http://www.nest-simulator.org/quickref/#nodes>`_
        for details).

    Returns
    -------
    recorders : list or NodeCollection of the recorders' gids
    recordables : list of the recordables' names.
    '''
    if nest_recorder is None:
        nest_recorder = [spike_rec]
    elif not nonstring_container(nest_recorder):
        nest_recorder = [nest_recorder]

    if params is None:
        params = [{}]
    elif isinstance(params, dict):
        params = [params]

    recorders   = NodeCollection([]) if nest_version == 3 else []
    recordables = []

    for name in group_names:
        gids = network.population[name].nest_gids

        recdr, recdbls = _monitor(gids, nest_recorder, params)

        recorders += recdr
        recordables.extend(recdbls)

    return recorders, recordables


def monitor_nodes(gids, nest_recorder=None, params=None, network=None):
    '''
    Monitoring the activity of nodes in the network.

    Parameters
    ----------
    gids : tuple of ints or list of tuples
        GIDs of the neurons in the NEST subnetwork; either one list per
        recorder if they should monitor different neurons or a unique list
        which will be monitored by all devices.
    nest_recorder : strings or list, optional (default: spike recorder)
        Device(s) to monitor the network.
    params : dict or list of, optional (default: `{}`)
        Dictionarie(s) containing the parameters for each recorder (see
        `NEST documentation <http://www.nest-simulator.org/quickref/#nodes>`_
        for details).
    network : :class:`~nngt.Network` or subclass, optional (default: None)
        Network which population will be used to differentiate groups.

    Returns
    -------
    recorders : list or NodeCollection containing the recorders' gids
    recordables : list of the recordables' names.
    '''
    if nest_recorder is None:
        nest_recorder = [spike_rec]
    elif not nonstring_container(nest_recorder):
        nest_recorder = [nest_recorder]

    if params is None:
        params = [{}]
    elif isinstance(params, dict):
        params = [params]

    return _monitor(gids, nest_recorder, params)


def _monitor(gids, nest_recorder, params):
    new_record = []
    recorders  = NodeCollection([]) if nest_version == 3 else []

    gids = _get_nest_gids(gids)

    for i, rec in enumerate(nest_recorder):
        # multi/volt/conductancemeter
        if "meter" in rec:
            device = None
            di_spec = {"rule": "all_to_all"}

            if not params[i].get("to_accumulator", True):
                device = nest.Create(rec, len(gids), _warn=False)
                di_spec["rule"] = "one_to_one"
            else:
                device = nest.Create(rec, _warn=False)

            recorders += device if nest_version == 3 else list(device)

            device_params = nest.GetDefaults(rec)
            device_params.update(params[i])
            new_record.append(device_params["record_from"])
            nest.SetStatus(device, params[i], _warn=False)
            nest.Connect(device, gids, conn_spec=di_spec, _warn=False)
        # event detectors
        elif rec == spike_rec:
            device = nest.Create(rec, _warn=False)

            recorders += device if nest_version == 3 else list(device)

            new_record.append("spikes")
            nest.SetStatus(device, params[i], _warn=False)
            nest.Connect(gids, device, _warn=False)
        else:
            raise InvalidArgument('Invalid recorder item in `nest_recorder`: '
                                  '{} is unknown.'.format(nest_recorder))
    return recorders, new_record


# ------------------- #
# Saving the activity #
# ------------------- #

def save_spikes(filename, recorder=None, network=None, save_positions=True,
                **kwargs):
    '''
    Plot the monitored activity.

    .. versionadded:: 0.7

    Parameters
    ----------
    filename : str
        Path to the file where the activity should be saved.
    recorder : tuple or list of tuples, optional (default: None)
        The NEST gids of the recording devices. If None, then all existing
        spike recorders are used.
    network : :class:`~nngt.Network` or subclass, optional (default: None)
        Network which activity will be monitored.
    save_positions : bool, optional (default: True)
        Whether to include the position of the neurons in the file; this
        requires `network` to be provided.
    **kwargs : see :func:`numpy.savetxt`
    '''
    rcrdrs = NodeCollection([]) if nest_version == 3 else []

    delim = kwargs.get('delimiter', ' ')

    if 'fmt' not in kwargs:
        kwargs['fmt'] = '%d{}%.6f'.format(delim)
    if 'header' not in kwargs:
        kwargs['header'] = 'Neuron{}Time'.format(delim)

    # normalize recorders and recordables
    if recorder is not None:
        for rec in recorder:
            if nest_version == 3:
                if isinstance(rec, NodeCollection):
                    rcrdrs += rec
                elif nonstring_container(rec):
                    rcrdrs += NodeCollection(rec)
                else:
                    rcrdrs += NodeCollection([rec])
            elif nonstring_container(rec):
                rcrdrs.append(rec)
            else:
                rcrdrs.append([rec])

            if nest_version == 3:
                if len(rcrdrs) == 1:
                    assert recrdrs.model == spike_rec, \
                        'Only spike_detectors are supported.'
                    assert recrdrs.model == (spike_rec,)*len(rcrdrs), \
                        'Only spike_detectors are supported.'
            else:
                assert (nest.GetStatus(rcrdrs, 'model')
                        == ('spike_detector',)*len(rcrdrs)), \
                       'Only spike_detectors are supported.'
    else:
        if nest_version == 3:
            rcrdrs = nest.GetNodes(properties={'model': spike_rec})
        else:
            rcrdrs = [[n] for n in nest.GetNodes(
                (0,), properties={'model': 'spike_detector'})[0]]

    if network is not None and network.is_spatial() and save_positions:
        save_positions = True
        kwargs['header'] += '{}X{}Y'.format(delim, delim)
        if delim in kwargs['fmt']:
            kwargs['fmt'] += '{}%.6f{}%.6f'.format(delim, delim)
    else:
        save_positions = False

    with open(filename, "wb") as f:
        for rec in rcrdrs:
            data = nest.GetStatus(rec, "events")[0]
            if len(data['senders']):
                if save_positions:
                    gids = np.unique(data['senders'])
                    gid_to_id = np.zeros(gids[-1] + 1, dtype=int)

                    for gid in gids:
                        gid_to_id[gid] = network.id_from_nest_gid(gid)

                    pos  = network.get_positions()
                    ids  = gid_to_id[data['senders']]
                    data = np.array(
                        (data['senders'], data['times'], pos[ids, 0],
                         pos[ids, 1])).T
                else:
                    data = np.array((data['senders'], data['times'])).T

                s = BytesIO()
                np.savetxt(s, data, **kwargs)
                f.write(s.getvalue())


def _get_nest_gids(gids):
    ''' Convert nodes to NodeCollection if NEST is version 3+ '''
    if nest_version == 3:
        if isinstance(gids, NodeCollection):
            return gids

        return NodeCollection(sorted(gids))

    if nonstring_container(gids):
        return list(gids)

    return [gids]



#!/usr/bin/env python
#-*- coding:utf-8 -*-

# test_generation.py

# This file is part of the NNGT module
# Distributed as a free software, in the hope that it will be useful, under the
# terms of the GNU General Public License.

"""
Test the main methods of the :mod:`~nngt.generation` module.
"""

import unittest
import numpy as np
import scipy.signal as sps

import nngt
from nngt.analysis import *
from nngt.lib.connect_tools import _compute_connections

from base_test import TestBasis, XmlHandler, network_dir
from tools_testing import foreach_graph


# ---------- #
# Test tools #
# ---------- #

def _get_connections(instruct):
    nodes = instruct["nodes"]
    density = instruct.get("density", -1.)
    edges = instruct.get("edges", -1)
    average_degree = instruct.get("avg_deg", -1.)
    reciprocity = instruct.get("reciprocity", -1.)
    directed = instruct.get("directed", True)
    #~ weighted = instruct.get("weighted", True))
    return nodes, density, edges, average_degree, directed, reciprocity


def _fixed_deg_theo(instruct):
    fixed_degree = instruct["degree"]
    avg_free_degree = fixed_degree
    deg_type = instruct["degree_type"]
    std_free_degree = np.sqrt(avg_free_degree) if deg_type != "total" else 0
    return fixed_degree, avg_free_degree, std_free_degree


def _fixed_deg_exp(graph, instruct):
    n, d, e, a_deg, directed, weighted = _get_connections(instruct)
    deg_type = instruct["degree_type"]
    deg_types = [ 'in', 'out' ]
    degrees = graph.get_degrees(deg_type)
    fixed_degree = degrees[0]
    # check that they are indeed all the same
    assert np.array_equal(degrees, np.full(len(degrees), fixed_degree))
    # compute free_degree properties
    avg_free_degree, std_free_degree = fixed_degree, 0
    if deg_type != 'total':
        del deg_types[deg_types.index(deg_type)]
        degrees = graph.get_degrees(deg_types[0], weights=False)
        avg_free_degree = np.average(degrees)
        std_free_degree = np.std(degrees)
    return fixed_degree, avg_free_degree, std_free_degree


def _gaussian_deg_theo(instruct):
    avg_degree = instruct["avg"]
    std_degree = instruct["std"]
    avg_free_degree = avg_degree
    deg_type = instruct["degree_type"]
    std_free_degree = np.sqrt(avg_free_degree)
    return avg_degree, std_degree, avg_free_degree, std_free_degree


def _gaussian_deg_exp(graph, instruct):
    n, d, e, a_deg, directed, weighted = _get_connections(instruct)
    deg_type = instruct["degree_type"]
    deg_types = [ 'in', 'out' ]
    degrees = graph.get_degrees(deg_type, weights=False)
    avg_degree = np.average(degrees)
    std_degree = np.std(degrees)
    # compute free_degree properties
    avg_free_degree, std_free_degree = avg_degree, std_degree
    if deg_type != 'total':
        del deg_types[deg_types.index(deg_type)]
        degrees = graph.get_degrees(deg_types[0], weights=False)
        avg_free_degree = np.average(degrees)
        std_free_degree = np.std(degrees)
    return avg_degree, std_degree, avg_free_degree, std_free_degree


def _erdos_renyi_theo(instruct):
    pass


def _erdos_renyi_exp(graph, instruct):
    pass


def _random_scale_free_theo(instruct):
    pass


def _random_scale_free_exp(graph, instruct):
    pass


def _newman_watts_theo(instruct):
    pass


def _newman_watts_exp(graph, instruct):
    pass


def _distance_rule_theo(instruct):
    # convention for distance rule:
    # - distribution is from 0 to 7*scale for exp, 0 to scale for lin
    # - bin size is 0.02*scale for exp, 0.005 for lin
    avg_deg = instruct["avg_deg"]
    res = [avg_deg]
    spatial_density = instruct["neuron_density"]
    scale = instruct["scale"]
    rule = instruct["rule"]
    dist_distrib = None
    distances = None
    if rule == 'exp':
        distances = np.arange(0.02*scale, 7*scale, 0.02*scale)
        def dist_distrib(d, space_dens, scale):
            fact = 2*np.pi*space_dens
            #~ max_val = fact*scale/np.e
            #~ norm = fact*scale**2
            return fact*d*np.exp(-d/scale)
    else:  # linear
        distances = np.arange(0.005*scale, scale, 0.005*scale)
        def dist_distrib(d, space_dens, scale):
            fact = 2*np.pi*space_dens
            #~ max_val = fact*scale/4.
            #~ norm = fact*scale**2 / 6.
            return fact*d*np.clip(scale-d, 0., np.inf) / scale
    distrib = dist_distrib(distances, spatial_density, scale)
    res.extend(distrib / distrib.sum())
    return res


def _distance_rule_exp(graph, instruct):
    # convention for distance rule:
    # - distribution is from 0 to 7*scale for exp, 0 to scale for lin
    # - bin size is 0.02*scale for exp, 0.005 for lin
    scale = instruct["scale"]
    rule = instruct["rule"]
    degrees = graph.get_degrees('out', weights=False)
    res = [np.average(degrees)]
    distances = graph.get_edge_attributes(name='distance')
    bins = None
    if rule == 'exp':
        bins = np.linspace(0, 7*scale, 350)
    else:
        bins = np.linspace(0, scale, 200)
    hist, _ = np.histogram(distances, bins)
    kernel = sps.gaussian(20, 3)
    hist = sps.convolve(hist, kernel, mode='same')
    res.extend(hist / hist.sum())
    return res


# ---------- #
# Test class #
# ---------- #

class TestGeneration(TestBasis):
    
    '''
    Class testing the main methods of the :mod:`~nngt.generation` module.
    '''
    
    theo_prop = {
        "fixed_degree": _fixed_deg_theo,
        "gaussian_degree": _gaussian_deg_theo,
        "erdos_renyi": _erdos_renyi_theo,
        "random_scale_free": _random_scale_free_theo,
        "newman_watts": _newman_watts_theo,
        "distance_rule": _distance_rule_theo,
    }
    
    exp_prop = {
        "fixed_degree": _fixed_deg_exp,
        "gaussian_degree": _gaussian_deg_exp,
        "erdos_renyi": _erdos_renyi_exp,
        "random_scale_free": _random_scale_free_exp,
        "newman_watts": _newman_watts_exp,
        "distance_rule": _distance_rule_exp,
    }

    tolerance = 0.08
    
    @property
    def test_name(self):
        return "test_generation"

    @unittest.skipIf(nngt.get_config('mpi'), 'Not checking for MPI')
    def gen_graph(self, graph_name):
        di_instructions = self.parser.get_graph_options(graph_name)
        graph = nngt.generate(di_instructions)
        graph.set_name(graph_name)
        return graph, di_instructions

    @foreach_graph
    def test_model_properties(self, graph, instructions, **kwargs):
        '''
        When generating graphs from on of the preconfigured models, check that
        the expected properties are indeed obtained.
        '''
        graph_type = instructions["graph_type"]
        ref_result = self.theo_prop[graph_type](instructions)
        computed_result = self.exp_prop[graph_type](graph, instructions)
        if graph_type == 'distance_rule':
            # average degree
            self.assertTrue(
                ref_result[0] == computed_result[0],
                "Avg. deg. for graph {} failed:\nref = {} vs exp {}\
                ".format(graph.name, ref_result[0], computed_result[0]))
            # average error on distance distribution
            sqd = np.square(np.subtract(ref_result[1:], computed_result[1:]))
            avg_sqd = sqd / np.square(computed_result[1:])
            err = np.sqrt(avg_sqd).mean()
            tolerance = (self.tolerance if instructions['rule'] == 'lin'
                         else 0.25)
            self.assertTrue(err <= tolerance,
                "Distance distribution for graph {} failed:\nerr = {} > {}\
                ".format(graph.name, err, tolerance))
        else:
            self.assertTrue(np.allclose(
                ref_result, computed_result, self.tolerance),
                "Test for graph {} failed:\nref = {} vs exp {}\
                ".format(graph.name, ref_result, computed_result))


# ---------- #
# Test suite #
# ---------- #

if not nngt.get_config('mpi'):
    suite = unittest.TestLoader().loadTestsFromTestCase(TestGeneration)

    if __name__ == "__main__":
        unittest.main()



import os, errno
from setuptools import setup, find_packages
from shutil import rmtree


# create directory
setup_dir = os.path.abspath(os.path.dirname(__file__))
directory = setup_dir + '/mpl_chord_diagram/'

try:
    os.makedirs(directory)
except OSError as e:
    if e.errno != errno.EEXIST:
        raise


from _info import __version__


# move important files
move = (
    '__init__.py',
    '_info.py',
    'LICENSE',
    'chord_diagram.py',
    'gradient.py',
    'utilities.py',
    'example.py',
)


for fname in move:
    try:
        os.rename(fname, directory + fname)
    except Exception as e:
        print(e)


long_descr = '''
The module enables to plot chord diagrams with matplotlib from lists,
numpy or scipy sparse matrices.
It provides tunable parameters such as arc sorting based on size or distance,
and supports gradients to better visualize source and target regions.
'''


try:
    # install
    setup(
        name = 'mpl_chord_diagram',
        version = __version__,
        description = 'Python module to plot chord diagrams with matplotlib.',
        package_dir = {'': '.'},
        packages = find_packages('.'),

        # Include the non python files:
        package_data = {'': ['LICENSE', '*.md']},

        # Requirements
        install_requires = ['numpy', 'scipy', 'matplotlib'],

        # Metadata
        url = 'https://codeberg.org/tfardet/mpl_chord_diagram',
        author = 'Tanguy Fardet',
        author_email = 'tanguy.fardet@tuebingen.mpg.de',
        license = 'MIT',
        keywords = 'chord diagram matplotlib plotting',
        long_description = long_descr,
        classifiers = [
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            'Operating System :: OS Independent',
            'Topic :: Scientific/Engineering :: Visualization',
            'Framework :: Matplotlib',
        ]
    )
except:
    pass
finally:
    for fname in move:
        try:
            os.rename(directory + fname, fname)
        except:
            pass

    try:
        rmtree(directory)
    except:
        pass



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Network generation for NEST '''

import numpy as np

import nngt
import nngt.generation as ng


# nngt.seed(0)


# -------------------- #
# Generate the network #
# -------------------- #

'''
Build a network with two populations:
* excitatory (80%)
* inhibitory (20%)
'''
num_nodes = 1000

# 800 excitatory neurons, 200 inhibitory
net = nngt.Network.exc_and_inhib(num_nodes, ei_ratio=0.2)

'''
Connect the populations.
'''
# exc -> inhib (Erdos-Renyi)
ng.connect_neural_types(net, 1, -1, "erdos_renyi", density=0.035)

# exc -> exc (Newmann-Watts)
prop_nw = {
    "coord_nb": 10,
    "proba_shortcut": 0.1,
    "reciprocity_circular": 1.
}
ng.connect_neural_types(net, 1, 1, "newman_watts", **prop_nw)

# inhib -> exc (Random scale-free)
prop_rsf = {
    "in_exp": 2.1,
    "out_exp": 2.6,
    "density": 0.2
}
ng.connect_neural_types(net, -1, 1, "random_scale_free", **prop_rsf)

# inhib -> inhib (Erdos-Renyi)
ng.connect_neural_types(net, -1, -1, "erdos_renyi", density=0.04)


# ------------------ #
# Simulate with NEST #
# ------------------ #

if nngt.get_config('with_nest'):
    import nest
    import nngt.simulation as ns

    '''
    Prepare the network and devices.
    '''
    # send to NEST
    gids = net.to_nest()
    # excite
    ns.set_poisson_input(gids, rate=100000.)
    # record
    groups = [key for key in net.population]
    recorder, record = ns.monitor_groups(groups, net)

    '''
    Simulate and plot.
    '''
    simtime = 100.
    nest.Simulate(simtime)

    if nngt.get_config('with_plot'):
        ns.plot_activity(
            recorder, record, network=net, show=True, limits=(0,simtime))

    '''
    A reminder about weights of inhibitory connections
    '''

    # sign of NNGT versus NEST inhibitory connections
    igroup = net.population["inhibitory"]

    # in NNGT
    iedges = net.get_edges(source_node=igroup.ids)
    w_nngt = set(net.get_weights(edges=iedges))

    # in NEST
    try:
        # nest 2
        iconn  = nest.GetConnections(
            source=list(net.population["inhibitory"].nest_gids),
            target=list(net.population.nest_gids))
    except:
        # nest 3
        import nest
        s = nest.NodeCollection(net.population["inhibitory"].nest_gids)
        t = nest.NodeCollection(net.population.nest_gids)

        iconn  = nest.GetConnections(source=s, target=t)

    w_nest = set(nest.GetStatus(iconn, "weight"))

    # In NNGT, inhibitory weights are positive to work with graph analysis
    # methods; they are automatically converted to negative weights in NEST
    print("NNGT weights:", w_nngt, "versus NEST weights", w_nest)

    assert w_nngt == {1} and w_nest == {-1}



#cython: cdivision=True, language_level=3
#!/usr/bin/env cython
#-*- coding:utf-8 -*-

""" Cython interface to C++ parallel generation tools for NNGT """

import warnings

from .cconnect cimport *
cimport numpy as cnp

import numpy as np
import scipy.sparse as ssp
from numpy.random import randint, get_state

import nngt
from nngt.lib import InvalidArgument
from nngt.lib.connect_tools import (_check_num_edges, _compute_connections,
                                    _set_degree_type, max_proba_dist_rule)


__all__ = [
    "_all_to_all",
    "_distance_rule",
    "_fixed_degree",
    "_from_degree_list",
    "_gaussian_degree",
]


cdef unsigned int MAXTESTS = 1000 # ensure that generation will finish
cdef float EPS = 0.00001

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
DTYPE = np.int64

ctypedef cnp.uint8_t uint8
ctypedef cnp.int64_t int64


# ------------ #
# Simple tools #
# ------------ #

cdef bytes _to_bytes(string):
    ''' Convert string to bytes '''
    if not isinstance(string, bytes):
        return bytes(string, "UTF-8")

    return string


def _no_self_loops(cnp.ndarray[int64_t, ndim=2] array):
    '''
    Remove self-loops
    '''
    cdef cnp.ndarray[bool, ndim=1] test = array[:, 0] != array[:, 1]

    return array[test], test


def _filter(cnp.ndarray[int64, ndim=2] ia_edges,
            cnp.ndarray[int64, ndim=2] ia_edges_tmp,
            size_t num_ecurrent, set edges_hash, bool b_one_pop,
            bool multigraph, bool directed=True, set recip_hash=None,
            cnp.ndarray distance=None, cnp.ndarray dist_tmp=None):
    '''
    Filter the edges: remove self loops and multiple connections if the graph
    is not a multigraph.
    '''
    if b_one_pop:
        ia_edges_tmp, test = _no_self_loops(ia_edges_tmp)

        if dist_tmp is not None:
            dist_tmp = dist_tmp[test]

    cdef:
        int64_t[:] e
        float d
        tuple tpl_e

    if not multigraph:
        num_ecurrent = len(edges_hash)

        if distance is not None:
            for e, d in zip(ia_edges_tmp, dist_tmp):
                tpl_e = tuple(e)

                if tpl_e not in edges_hash:
                    if directed or tpl_e not in recip_hash:
                        ia_edges[num_ecurrent] = e
                        distance.append(d)
                        edges_hash.add(tpl_e)

                        if not directed:
                            recip_hash.add(tpl_e[::-1])

                        num_ecurrent += 1
        else:
            for e in ia_edges_tmp:
                tpl_e = tuple(e)

                if tpl_e not in edges_hash:
                    if directed or tpl_e not in recip_hash:
                        ia_edges[num_ecurrent] = e
                        edges_hash.add(tpl_e)

                        if not directed:
                            recip_hash.add(tpl_e[::-1])

                        num_ecurrent += 1
    else:
        num_added = len(ia_edges_tmp)
        ia_edges[num_ecurrent:num_ecurrent + num_added, :] = ia_edges_tmp
        num_ecurrent += num_added

        if distance is not None:
            distance.extend(dist_tmp)

    return ia_edges, num_ecurrent


# ---------------------- #
# Graph model generation #
# ---------------------- #

def _all_to_all(cnp.ndarray[size_t, ndim=1] source_ids,
                cnp.ndarray[size_t, ndim=1] target_ids,
                bool directed=True, bool multigraph=False,
                list distance=None, **kwargs):
    ''' Generation of a fully connected graph '''
    cdef:
        size_t num_sources = len(source_ids)
        size_t num_targets = len(target_ids)
        cnp.ndarray[float, ndim=2, mode="c"] vectors
        cnp.ndarray[float, ndim=1] x, y
        size_t s, next_enum

    # find common nodes
    edges  = None
    common = set(source_ids).intersection(target_ids)

    num_edges = int(0.5*(1 + directed)*num_sources*num_targets)

    if common:
        num_edges -= int(0.5*(1 + directed)*len(common))

        edges         = np.full((num_edges, 2), -1, dtype=DTYPE)
        current_edges = 0
        next_enum     = 0

        for s in source_ids:
            if s in common:
                if directed:
                    next_enum = current_edges + num_targets - 1

                    edges[current_edges:next_enum, 1] = \
                        target_ids[target_ids != s]
                else:
                    tgts = [t for t in target_ids if t not in common or t > s]

                    next_enum = current_edges + len(tgts)

                    edges[current_edges:next_enum, 1] = tgts

                edges[current_edges:next_enum, 0] = s

                current_edges = next_enum
            else:
                next_enum = current_edges + num_targets
                edges[current_edges:next_enum, 0] = s
                edges[current_edges:next_enum, 1] = target_ids
                current_edges = next_enum
    else:
        edges       = np.empty((num_sources*num_targets, 2), dtype=DTYPE)
        edges[:, 0] = np.repeat(source_ids, num_targets)
        edges[:, 1] = np.tile(target_ids, num_sources)

    if distance is not None and 'positions' in kwargs:
        pos       = kwargs['positions']
        x, y      = pos[0], pos[1]
        vectors   = np.array((x[edges[:, 1]] - x[edges[:, 0]],
                              y[edges[:, 1]] - y[edges[:, 0]]))
        distance.extend(np.linalg.norm(vectors, axis=0))

    return edges


def _total_degree_list(cnp.ndarray[int64, ndim=1] source_ids,
                       cnp.ndarray[int64, ndim=1] target_ids,
                       cnp.ndarray[int64, ndim=1] degree_list,
                       bool directed=True, bool multigraph=False, **kwargs):
    ''' Called from _from_degree_list '''
    cdef:
        size_t num_source = len(source_ids)
        size_t num_target = len(target_ids)
        size_t edges = 0.5*np.sum(degree_list)

    # check if the sequence is not obviously impossible
    if not multigraph:
        assert np.all(np.less_equal(degree_list, (num_target - 1))), \
            "Some degrees are higher than the number of available targets."

    if np.sum(degree_list) % 2 != 0:
        raise ValueError("The sum of the degrees must ben even.")

    # check one pop
    cdef:
        bool b_one_pop
        set source_set, target_set

    b_one_pop, source_set, target_set = _check_num_edges(
        source_ids, target_ids, edges, directed, multigraph, return_sets=True)

    cdef:
        size_t i, k, num_tests, num_choice, new_etot
        size_t ecurrent = 0
        set edges_hash = set()
        set recip_hash = set()
        set ids
        int64[:] sorted_degrees = np.sort(degree_list)[::-1]
        int64[:] data
        cnp.ndarray[int64, ndim=1] incr, decr
        cnp.ndarray[int64, ndim=1] deg_tmp, unfinished, even_deg, add_node
        cnp.ndarray[int64, ndim=2] chosen, new_edges
        cnp.ndarray[int64, ndim=2] ia_edges = np.full((edges, 2), -1,
                                                      dtype=np.int64)

    rng = nngt._rng

    if b_one_pop:
        # check that the degree sequence is graphic (i.e. can be realized)
        for k in range(1, num_source + 1):
            partial_sum = np.sum(sorted_degrees[:k])
            capped_deg  = np.minimum(sorted_degrees[k:], k)

            if not partial_sum <= k*(k-1) + np.sum(capped_deg):
                raise ValueError("The degree sequence provided is not "
                                 "graphical and cannot be realized.")

        num_tests = 0

        while ecurrent < edges and num_tests < MAXTESTS:
            # initial procedure
            remaining = max(int(0.5*(edges - ecurrent)), 1)

            data = np.repeat(source_ids, degree_list)
            sources = rng.choice(data, remaining, replace=False)

            deg_tmp = degree_list.copy()
            np.add.at(deg_tmp, sources, -1)

            data = np.repeat(source_ids, deg_tmp)
            targets = rng.choice(data, remaining, replace=False)

            new_edges = np.array((sources, targets), dtype=DTYPE).T

            ia_edges, new_etot =  _filter(
                ia_edges, new_edges, ecurrent, edges_hash, b_one_pop,
                multigraph, directed=directed, recip_hash=recip_hash)

            np.add.at(degree_list, np.ravel(ia_edges[ecurrent:new_etot]), -1)

            ecurrent = new_etot

            # trying to correct things if initial procedure did not converge
            if num_tests >= 0.5*MAXTESTS:
                nnz_degree = degree_list != 0
                unfinished = np.where(nnz_degree)[0]
                num_choice = max(int(0.5*len(unfinished)), 1)

                # pick random edges
                ids    = set(rng.choice(ecurrent, num_choice, replace=False))
                chosen = np.array([ia_edges[i] for i in ids], dtype=np.int64)

                # if unfinished is odd then we need to readd a node that has
                # a remaining degree of 2
                if len(unfinished) % 2 == 1:
                    even_deg = np.where((degree_list % 2 == 0)*nnz_degree)[0]
                    add_node = rng.choice(even_deg, 1)
                    unfinished = np.array(
                        list(unfinished) + list(add_node), dtype=np.int64)

                # try to pair them differently
                rng.shuffle(unfinished)

                new_edges = np.array(
                    [(u, v) for u, v in zip(unfinished, chosen.ravel())],
                    dtype=np.int64)

                # randomize direction if directed
                if directed:
                    num_reverse = rng.binomial(2*num_choice, 0.5)
                    reverse     = rng.choice(2*num_choice, num_reverse,
                                             replace=False)

                    new_edges[reverse, 0], new_edges[reverse, 1] = \
                        new_edges[reverse, 1], new_edges[reverse, 0]

                # check that new_edges are indeed new
                skip = False

                for e in new_edges:
                    if e[0] == e[1]:
                        skip = True
                        break

                    tpl_e = tuple(e)
                        
                    if tpl_e in edges_hash:
                        skip = True
                        break

                    if not directed and tpl_e in recip_hash:
                        skip = True
                        break

                if skip:
                    num_tests += 1
                    continue

                # remove old ones from edges_hash
                edges_hash -= {tuple(e) for e in chosen}

                if not directed:
                    recip_hash -= {tuple(e) for e in chosen[:, ::-1]}

                # remove chosen edges from existing edges
                ia_edges[:ecurrent - num_choice] = np.array(
                    [e for i, e in enumerate(ia_edges[:ecurrent])
                     if i not in ids])

                ecurrent -= num_choice

                ia_edges, new_etot =  _filter(
                    ia_edges, new_edges, ecurrent, edges_hash, b_one_pop,
                    multigraph, directed=directed, recip_hash=recip_hash)

                decr  = new_edges.ravel()
                incr  = chosen.ravel()

                np.add.at(degree_list, incr, 1)
                np.add.at(degree_list, decr, -1)
                
                ecurrent += len(new_edges)
                
            num_tests += 1

        if num_tests == MAXTESTS:
            raise RuntimeError("Graph generation did not converge.")

        return ia_edges

    raise NotImplementedError("not available if sources != targets.")


def _from_degree_list(cnp.ndarray[size_t, ndim=1] source_ids,
                      cnp.ndarray[size_t, ndim=1] target_ids, degrees,
                      degree_type="in", bool directed=True,
                      bool multigraph=False, existing_edges=None, **kwargs):
    ''' Generation of the degree list through the C++ function '''

    assert len(degrees) == len(source_ids), \
        "One degree per source neuron must be provided."

    degree_type = _set_degree_type(degree_type)

    cdef:
        # type of degree
        bool b_out = (degree_type == "out")
        bool b_total = (degree_type == "total")
        bool use_directed = directed + b_total
        size_t num_source = source_ids.shape[0]
        size_t edges = np.sum(degrees)
        bool b_one_pop = _check_num_edges(
            source_ids, target_ids, edges, directed, multigraph)
        int64_t[:, :] ia_edges = np.full((edges, 2), -1, dtype=DTYPE)

        vector[size_t] sources = source_ids
        vector[size_t] targets = target_ids

        unsigned int idx = 0 if b_out else 1  # differenciate source / target
        unsigned int omp = nngt._config["omp"]
        cnp.ndarray[int64, ndim=1] degree_list, source64, target64
        vector[unsigned int] cdegrees = degrees
        vector[ vector[size_t] ] old_edges = vector[ vector[size_t] ]()
        vector[long] seeds = _random_init(omp)

    # total-degree or undirected case
    if b_total or not directed:
        degree_list = np.array(degrees, dtype=np.int64)
        source64 = np.array(source_ids, dtype=np.int64)
        target64 = np.array(target_ids, dtype=np.int64)

        return _total_degree_list(source64, target64, degree_list,
                                  directed=directed, multigraph=multigraph)

    if existing_edges is not None:
        old_edges.push_back(list(existing_edges[:, 0]))
        old_edges.push_back(list(existing_edges[:, 1]))

    # directed case for in/out-degrees
    _gen_edges(&ia_edges[0,0], source_ids, cdegrees, target_ids, old_edges,
               idx, multigraph, use_directed, seeds)

    return ia_edges


def _fixed_degree(cnp.ndarray[size_t, ndim=1] source_ids,
                  cnp.ndarray[size_t, ndim=1] target_ids, degree=-1,
                  degree_type="in", float reciprocity=-1, bool directed=True,
                  bool multigraph=False, existing_edges=None, **kwargs):
    ''' Generation of the edges through the C++ function '''
    degree = int(degree)

    assert degree >= 0, "A positive value is required for `degree`."

    degrees = np.repeat(degree, len(source_ids))

    return _from_degree_list(source_ids, target_ids, degrees,
                             degree_type=degree_type, directed=directed,
                             multigraph=multigraph,
                             existing_edges=existing_edges)


def _gaussian_degree(cnp.ndarray[size_t, ndim=1] source_ids,
                     cnp.ndarray[size_t, ndim=1] target_ids, float avg=-1,
                     float std=-1, degree_type="in", float reciprocity=-1,
                     bool directed=True, bool multigraph=False,
                     existing_edges=None, **kwargs):
    '''
    Connect nodes with a Gaussian distribution (generation through C++
    function.
    '''
    # switch values to float
    avg = float(avg)
    std = float(std)

    assert avg >= 0, "A positive value is required for `avg`."
    assert std >= 0, "A positive value is required for `std`."

    degree_type = _set_degree_type(degree_type)

    rng = nngt._rng

    cdef:
        # type of degree
        bool b_total = (degree_type == "total")
        size_t num_source = source_ids.shape[0]
        uint64_t[:] degrees = np.around(np.maximum(
            rng.normal(avg, std, num_source), 0)).astype(np.uint64)
        size_t edges = np.sum(degrees)
        int idx

    if b_total or not directed:
        # check that the sum of the degrees is even
        if edges % 2 != 0:
            idx = -1

            # correct if its not the case
            while idx < 0 or degrees[idx] == 0:
                idx = rng.choice(num_source)
                if degrees[idx] > 0:
                    degrees[idx] -= 1

    return _from_degree_list(source_ids, target_ids, degrees,
        degree_type=degree_type, directed=directed, multigraph=multigraph,
        existing_edges=existing_edges)


def _distance_rule(cnp.ndarray[size_t, ndim=1] source_ids,
                   cnp.ndarray[size_t, ndim=1] target_ids, density=None,
                   edges=None, avg_deg=None, float scale=-1., str rule="exp",
                   float max_proba=-1., shape=None,
                   cnp.ndarray[float, ndim=2] positions=np.array([[0], [0]]),
                   bool directed=True, bool multigraph=False,
                   num_neurons=None, distance=None, **kwargs):
    '''
    Returns a distance-rule graph
    '''
    distance = [] if distance is None else distance

    if num_neurons is None:
        num_neurons = len(set(np.concatenate((source_ids, target_ids))))

    cdef:
        size_t cnum_neurons = num_neurons
        size_t num_source = source_ids.shape[0]
        size_t num_target = target_ids.shape[0]
        size_t s, i, edge_num
        string crule = _to_bytes(rule)
        unsigned int omp = nngt._config["omp"]
        vector[ vector[size_t] ] old_edges = vector[ vector[size_t] ]()
        vector[ vector[size_t] ] local_targets
        cnp.ndarray[long, ndim=1] loc_tgts
        list sources = []
        list targets = []
        vector[float] x = positions[0]
        vector[float] y = positions[1]
        float cscale = scale

    # compute the required values
    if max_proba <= 0.:
        edge_num, _ = _compute_connections(
            num_source, num_target, density, edges, avg_deg, directed)
    else:
        raise RuntimeError(
            "Not working with max_proba and multithreading yet.")

    existing = 0  # for now
    b_one_pop = _check_num_edges(
        source_ids, target_ids, edge_num, directed, multigraph)

    # for each node, check the neighbours that are in an area where
    # connections can be made: +/- scale for lin, +/- 10*scale for exp
    lim = scale if rule == 'lin' else 10*scale
    for i in source_ids:
        keep  = (np.abs(positions[0, target_ids] - positions[0, i]) < lim)
        keep *= (np.abs(positions[1, target_ids] - positions[1, i]) < lim)
        if b_one_pop:
            idx = np.where(target_ids == i)[0][0]
            keep[idx] = 0
        local_targets.push_back(target_ids[keep].tolist())

    # create the edges
    cdef:
        size_t cedges = edge_num
        int64_t[:, :] ia_edges = np.full((existing + edge_num, 2), -1,
                                         dtype=DTYPE)
        vector[float] dist = vector[float]()
        vector[long] seeds = _random_init(omp)

    if max_proba <= 0.:
        _cdistance_rule(&ia_edges[0,0], source_ids, local_targets, crule,
                        cscale, 1., x, y, cnum_neurons, cedges, old_edges,
                        dist, multigraph, directed, seeds)
        distance.extend(dist)
        return ia_edges

    for i, s in enumerate(source_ids):
        loc_tgts = np.array(local_targets[i], dtype=int)
        if len(loc_tgts):
            dist_tmp = []
            test = max_proba_dist_rule(
                rule, scale, max_proba, positions[:, s],
                positions[:, loc_tgts], dist=dist_tmp)
            test = np.greater(test, np.random.uniform(size=len(test)))
            added = np.sum(test)
            sources.extend((s for i in range(added)))
            targets.extend(loc_tgts[test])
            distance.extend(np.array(dist_tmp)[test])
    return np.array([sources, targets]).T


# ------------ #
# Random seeds #
# ------------ #


def _random_init(omp):
    '''
    Init the local random seeds
    '''
    # compute local random seeds
    seeds = None

    if not nngt._seeded_local:
        # generate the local seeds for the first time
        msd   = nngt._config["msd"]
        seeds = [msd + i + 1 for i in range(omp)]

        nngt._config["seeds"] = seeds
    elif nngt._used_local:
        # the initial seeds have already been used so we generate new ones
        msd   = np.random.randint(0, 2**31 - omp - 1)
        seeds = [msd + i + 1 for i in range(omp)]
    else:
        seeds = nngt.get_config('seeds')

    assert len(seeds) == omp, "Wrong number of seeds, need one per thread."

    nngt._seeded_local = True
    nngt._used_local   = True

    return seeds



#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2020  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Hive plot panel
===============
"""

import os

import matplotlib.pyplot as plt

import nngt


plt.rcParams.update({
    "figure.facecolor": (0, 0, 0, 0),
    "axes.titlecolor": "grey", "text.color": "grey"
})


dirpath  = os.path.abspath(os.getcwd())
rootpath = os.path.abspath(dirpath + "/../../..")


# load graph

g = nngt.load_from_file(rootpath + "/testing/Networks/rat_brain.graphml",
                        attributes=["weight"], cleanup=True,
                        attributes_types={"weight": float})


# prepare attributes

cc = nngt.analysis.local_clustering(g, weights="weight")

g.new_node_attribute("cc", "double", values=cc)

g.new_node_attribute("strength", "double",
                     values=g.get_degrees(weights="weight"))

flux = g.get_degrees("out") - g.get_degrees("in")

g.new_node_attribute("flux", "double", values=flux)


# figure parameters

cc_bins = [0, 0.1, 0.25, 0.6]

todo = ["strength", "cc", "flux"]
bins = [3, cc_bins, 3]


# make plot

fig, axes = plt.subplots(len(todo), len(todo), figsize=(10, 9))

for i in range(len(todo)):
    radial = todo[i]

    for j in range(len(todo)):
        ax_name = todo[j]
        ax_bins = bins[j]

        ax = axes[i, j]

        if i == 0:
            ax.set_title(ax_name)

        if j == 0:
            ax.set_ylabel(radial)

        size = todo[list(set([0, 1, 2]).difference([i, j]))[0]]

        nngt.plot.hive_plot(
            g, radial, axes=ax_name, edge_alpha=0.1, nsize=size, max_nsize=50,
            axes_bins=ax_bins, axes_units="native", axis=ax, show_names=False)

for i in range(len(todo)):
    fig.text(0.03, 0.8 - i*0.33, todo[i], rotation=90, fontsize="large")

plt.tight_layout()

plt.show()



#-*- coding:utf-8 -*-
#
# lib/db_tools.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" Database tools for NNGT """

import platform as pfm
from collections import namedtuple


try:
    import psutil
except ImportError:
    
    class MockPsutil:
        ''' Mock class for psutil '''
        
        @staticmethod
        def cpu_count():
            return -1

        @staticmethod
        def virtual_memory():
            attr = ('total', 'available', 'percent', 'used', 'free', 'active',
                    'inactive', 'buffers', 'cached')
            VM = namedtuple('VM', attr)
            return VM( ( -1 for _ in attr ) )

    psutil = MockPsutil



#-*- coding:utf-8 -*-
#
# simulation/__init__.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Module to interact easily with the NEST simulator. It allows to:

* build a NEST network from :class:`~nngt.Network` or
  :class:`~nngt.SpatialNetwork` objects,
* monitor the activity of the network (taking neural groups into account)
* plot the activity while separating the behaviours of predefined neural groups
"""

import sys as _sys
import logging as _logging

import nngt as _nngt
from nngt.lib.logger import _log_message


_logger = _logging.getLogger(__name__)


# -------------- #
# Import modules #
# -------------- #

from . import nest_activity as _na
from . import nest_graph as _ng
from . import nest_utils as _nu
from .nest_activity import *
from .nest_graph import *
from .nest_utils import *


# ----------------- #
# Declare functions #
# ----------------- #

__all__ = []
__all__.extend(_na.__all__)
__all__.extend(_ng.__all__)
__all__.extend(_nu.__all__)

# test import of simulation plotting tools

try:
    import matplotlib
    _with_plot = True
except ImportError:
    _with_plot = False

if _with_plot:
    from .nest_plot import plot_activity, raster_plot
    __all__.extend(("plot_activity", "raster_plot"))


# ---------------- #
# Wrap ResetKernel #
# ---------------- #

from nest import ResetKernel as _rk
from nest import Connect as _conn
from nest import Disconnect as _disc
from nest import Create as _cr
from nest import SetStatus as _setstat

try:
    from nest import ResetNetwork as _rn
except ImportError:
    pass

# store old functions
if not _nngt._old_nest_func:
    _nngt._old_nest_func["ResetKernel"] = _rk
    _nngt._old_nest_func["Connect"]     = _conn
    _nngt._old_nest_func["Disconnect"]  = _disc
    _nngt._old_nest_func["Create"]      = _cr
    _nngt._old_nest_func["SetStatus"]   = _setstat

    try:
        _nngt._old_nest_func["ResetNetwork"] = _rn
    except NameError:
        pass
else:
    _rk      = _nngt._old_nest_func["ResetKernel"]
    _conn    = _nngt._old_nest_func["Connect"]
    _disc    = _nngt._old_nest_func["Disconnect"]
    _cr      = _nngt._old_nest_func["Create"]
    _setstat = _nngt._old_nest_func["SetStatus"]

    try:
        _rn = _nngt._old_nest_func["ResetNetwork"]
    except KeyError:
        pass


def _new_reset_kernel():
    '''
    Call nest.ResetKernel, then reset all NeuralPops and parent Networks.
    '''
    _rk()
    _nngt.NeuralPop._nest_reset()


def _new_nest_func(old_nest_func):
    '''
    Print a warning to make sure user know what they are doing.
    '''
    def wrapper(*args, **kwargs):
        if kwargs.get("_warn", True):
            _log_message(_logger, "WARNING", "This function could interfere "
                         "with NNGT, making your Network obsolete compared to "
                         "the one in NEST... make sure to check what is "
                         "modified!")

        if "_warn" in kwargs:
            del kwargs["_warn"]

        return old_nest_func(*args, **kwargs)

    return wrapper


# nest is in sysmodules because it was imported in the main __init__.py
_sys.modules["nest"].ResetKernel = _new_reset_kernel
_sys.modules["nest"].Connect     = _new_nest_func(_conn)
_sys.modules["nest"].Disconnect  = _new_nest_func(_disc)
_sys.modules["nest"].Create      = _new_nest_func(_cr)
_sys.modules["nest"].SetStatus   = _new_nest_func(_setstat)

try:
    _sys.modules["nest"].ResetNetwork = _new_nest_func(_rn)
except NameError:
    pass



#-*- coding:utf-8 -*-
#
# core/ig_graph.py
#
# This file is part of the NNGT project, a graph-library for standardized and
# and reproducible graph analysis: generate and analyze networks with your
# favorite graph library (graph-tool/igraph/networkx) on any platform, without
# any change to your code.
# Copyright (C) 2015-2021 Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

""" iGraph subclassing """

from collections import OrderedDict
from copy import deepcopy
import logging

import numpy as np
import scipy.sparse as ssp

import nngt
from nngt.lib import InvalidArgument, nonstring_container, BWEIGHT, is_integer
from nngt.lib.connect_tools import (_cleanup_edges, _set_dist_new_edges,
                                    _set_default_edge_attributes)
from nngt.lib.graph_helpers import (_get_dtype, _get_ig_weights,
                                    _post_del_update)
from nngt.lib.converters import _np_dtype, _to_np_array
from nngt.lib.logger import _log_message
from .graph_interface import GraphInterface, BaseProperty


logger = logging.getLogger(__name__)


# ---------- #
# Properties #
# ---------- #

class _IgNProperty(BaseProperty):

    '''
    Class for generic interactions with nodes properties (igraph)
    '''

    def __getitem__(self, name):
        g = self.parent()._graph
        dtype = _np_dtype(super(_IgNProperty, self).__getitem__(name))
        return _to_np_array(g.vs[name], dtype=dtype)

    def __setitem__(self, name, value):
        g    = self.parent()._graph
        size = g.vcount()

        if name in self:
            if len(value) == size:
                g.vs[name] = value
            else:
                raise ValueError("A list or a np.array with one entry per "
                                 "node in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

    def new_attribute(self, name, value_type, values=None, val=None):
        g = self.parent()._graph

        if val is None:
            if value_type == "int":
                val = int(0)
            elif value_type == "double":
                val = np.NaN
            elif value_type == "string":
                val = ""
            else:
                val = None
                value_type = "object"

        if values is None:
            values = [deepcopy(val) for _ in range(g.vcount())]

        # store name and value type in the dict
        super(_IgNProperty,self).__setitem__(name, value_type)
        # store the real values in the attribute
        self[name] = values
        self._num_values_set[name] = len(values)

    def set_attribute(self, name, values, nodes=None):
        '''
        Set the node attribute.

        Parameters
        ----------
        name : str
            Name of the node attribute.
        values : array, size N
            Values that should be set.
        nodes : array-like, optional (default: all nodes)
            Nodes for which the value of the property should be set. If `nodes`
            is not None, it must be an array of size N.
        '''
        g = self.parent()._graph

        num_nodes = g.vcount()
        num_n = len(nodes) if nodes is not None else num_nodes

        if num_n == num_nodes:
            self[name] = values
        elif num_n:
            if num_n != len(values):
                raise ValueError("`nodes` and `values` must have the same "
                                 "size; got respectively " + str(num_n) + \
                                 " and " + str(len(values)) + " entries.")
            if self._num_values_set[name] == num_nodes - num_n:
                g.vs[-num_n:][name] = values
            else:
                for n, val in zip(nodes, values):
                    g.vs[n][name] = val
        if num_n:
            self._num_values_set[name] = num_nodes


class _IgEProperty(BaseProperty):

    ''' Class for generic interactions with nodes properties (igraph) '''

    def __getitem__(self, name):
        g = self.parent()._graph

        if isinstance(name, slice):
            eprop = {}
            for k in self.keys():
                dtype = _np_dtype(super(_IgEProperty, self).__getitem__(k))
                eprop[k] = _to_np_array(g.es[k], dtype=dtype)[name]
            return eprop
        elif nonstring_container(name):
            eprop = {}
            if nonstring_container(name[0]):
                eids = [g.get_eid(*e) for e in name]
                for k in self.keys():
                    dtype = _np_dtype(super(_IgEProperty, self).__getitem__(k))
                    eprop[k] = _to_np_array(g.es[k], dtype=dtype)[eids]
            else:
                eid = g.get_eid(*name)
                for k in self.keys():
                    eprop[k] = g.es[k][eid]
            return eprop

        dtype = _np_dtype(super(_IgEProperty, self).__getitem__(name))

        return _to_np_array(g.es[name], dtype=dtype)

    def __setitem__(self, name, value):
        g = self.parent()._graph

        if name in self:
            size = g.ecount()
            if len(value) == size:
                g.es[name] = value
            else:
                raise ValueError("A list or a np.array with one entry per "
                                 "edge in the graph is required")
        else:
            raise InvalidArgument("Attribute does not exist yet, use "
                                  "set_attribute to create it.")

    def new_attribute(self, name, value_type, values=None, val=None):
        g = self.parent()._graph

        if val is None:
            if value_type == "int":
                val = int(0)
            elif value_type == "double":
                val = np.NaN
            elif value_type == "string":
                val = ""
            else:
                val = None
                value_type = 'object'

        if values is None:
            values = [deepcopy(val) for _ in range(g.ecount())]

        # store name and value type in the dict
        super(_IgEProperty, self).__setitem__(name, value_type)
        # store the real values in the attribute
        self[name] = values
        self._num_values_set[name] = len(values)

    def set_attribute(self, name, values, edges=None, last_edges=False):
        '''
        Set the edge property.

        Parameters
        ----------
        name : str
            Name of the edge property.
        values : array
            Values that should be set.
        edges : array-like, optional (default: None)
            Edges for which the value of the property should be set. If `edges`
            is not None, it must be an array of shape `(len(values), 2)`.
        '''
        g = self.parent()._graph

        num_edges = g.ecount()
        num_e = len(edges) if edges is not None else num_edges

        if num_e != len(values):
            raise ValueError(
                "`edges` and `values` must have the same  size; got "
                "respectively " + str(num_e) + " and " + str(len(values))
                + " entries.")

        if edges is None:
            self[name] = values
        else:
            if last_edges:
                g.es[-num_e:][name] = values
            else:
                for e, val in zip(edges, values):
                    eid = g.get_eid(*e)
                    g.es[eid][name] = val

                self._num_values_set[name] += num_e

        if num_e:
            self._num_values_set[name] = num_edges


# ----- #
# Graph #
# ----- #

class _IGraph(GraphInterface):

    '''
    Subclass of :class:`igraph.Graph`.
    '''

    _nattr_class = _IgNProperty
    _eattr_class = _IgEProperty    

    #-------------------------------------------------------------------------#
    # Constructor and instance properties

    def __init__(self, nodes=0, copy_graph=None, directed=True, weighted=False,
                 **kwargs):
        self._nattr = _IgNProperty(self)
        self._eattr = _IgEProperty(self)

        g = copy_graph.graph if copy_graph is not None else None

        if g is None:
            self._graph = nngt._config["graph"](n=nodes, directed=directed)
        else:
            # convert graph if necessary
            if directed and not g.is_directed():
                g = g.copy()
                g.to_directed()
            elif not directed and g.is_directed():
                g = g.as_undirected(mode="collapse", combine_edges="sum")
                g.simplify(combine_edges="sum")

            self._from_library_graph(g, copy=True)

    #-------------------------------------------------------------------------#
    # Graph manipulation

    def edge_id(self, edge):
        '''
        Return the ID a given edge or a list of edges in the graph.
        Raises an error if the edge is not in the graph or if one of the
        vertices in the edge is nonexistent.

        Parameters
        ----------
        edge : 2-tuple or array of edges
            Edge descriptor (source, target).

        Returns
        -------
        index : int or array of ints
            Index of the given `edge`.
        '''
        if is_integer(edge[0]):
            return self._graph.get_eid(*edge)
        elif nonstring_container(edge[0]):
            return self._graph.get_eids(edge)
        else:
            raise AttributeError("`edge` must be either a 2-tuple of ints or "
                                 "an array of 2-tuples of ints.")

    def has_edge(self, edge):
        '''
        Whether `edge` is present in the graph.

        .. versionadded:: 2.0
        '''
        try:
            self._graph.get_eid(*edge)
            return True
        except:
            return False

    @property
    def edges_array(self):
        '''
        Edges of the graph, sorted by order of creation, as an array of
        2-tuple.
        '''
        g = self._graph
        return np.array([e.tuple for e in g.es], dtype=int)

    def _get_edges(self, source_node=None, target_node=None):
        '''
        Called by Graph.get_edges if source_node and target_node are not both
        integers.
        '''
        g = self._graph

        edges = None

        if source_node is None:
            if target_node is None:
                edges = g.es
            elif is_integer(target_node):
                edges = g.es.select(_target_eq=target_node)
            else:
                edges = g.es.select(_target_in=target_node)
        elif is_integer(source_node):
            if target_node is None:
                edges = g.es.select(_source_eq=source_node)
            else:
                edges = g.es.select(_source_eq=source_node,
                                    _target_in=target_node)
        else:
            if target_node is None:
                edges = g.es.select(_source_in=source_node)
            elif is_integer(target_node):
                edges = g.es.select(_source_in=source_node,
                                    _target_eq=target_node)
            else:
                edges = g.es.select(_source_in=source_node,
                                    _target_in=target_node)

        return [e.tuple for e in edges]

    def new_node(self, n=1, neuron_type=1, attributes=None, value_types=None,
                 positions=None, groups=None):
        '''
        Adding a node to the graph, with optional properties.

        Parameters
        ----------
        n : int, optional (default: 1)
            Number of nodes to add.
        neuron_type : int, optional (default: 1)
            Type of neuron (1 for excitatory, -1 for inhibitory)
        attributes : dict, optional (default: None)
            Dictionary containing the attributes of the nodes.
        value_types : dict, optional (default: None)
            Dict of the `attributes` types, necessary only if the `attributes`
            do not exist yet.
        positions : array of shape (n, 2), optional (default: None)
            Positions of the neurons. Valid only for
            :class:`~nngt.SpatialGraph` or :class:`~nngt.SpatialNetwork`.
        groups : str, int, or list, optional (default: None)
            :class:`~nngt.core.NeuralGroup` to which the neurons belong. Valid
            only for :class:`~nngt.Network` or :class:`~nngt.SpatialNetwork`.

        Returns
        -------
        The node or an iterator over the nodes created.
        '''
        g = self._graph

        first_node_idx = g.vcount()

        g.add_vertices(n)

        nodes = list(range(first_node_idx, first_node_idx + n))

        attributes = {} if attributes is None else deepcopy(attributes)

        if attributes:
            for k, v in attributes.items():
                if k not in self._nattr:
                    self._nattr.new_attribute(k, value_types[k], val=v)
                else:
                    v = v if nonstring_container(v) else [v]
                    self._nattr.set_attribute(k, v, nodes=nodes)

        # set default values for all attributes that were not set
        for k in self.node_attributes:
            if k not in attributes:
                dtype = self.get_attribute_type(k)
                if dtype == "string":
                    self._nattr.set_attribute(k, ["" for _ in nodes],
                                              nodes=nodes)
                elif dtype == "int":
                    self._nattr.set_attribute(k, [0 for _ in nodes],
                                              nodes=nodes)
                elif dtype == "double":
                    self._nattr.set_attribute(k, [np.NaN for _ in nodes],
                                              nodes=nodes)

        g.vs[nodes[0]:nodes[-1] + 1]['type'] = neuron_type

        if self.is_spatial():
            old_pos      = self._pos
            self._pos    = np.full((self.node_nb(), 2), np.NaN)
            num_existing = len(old_pos) if old_pos is not None else 0
            if num_existing != 0:
                self._pos[:num_existing, :] = old_pos
        if positions is not None:
            assert self.is_spatial(), \
                "`positions` argument requires a SpatialGraph/SpatialNetwork."
            self._pos[nodes] = positions

        if groups is not None:
            assert self.is_network(), \
                "`positions` argument requires a Network/SpatialNetwork."
            if nonstring_container(groups):
                assert len(groups) == n, "One group per neuron required."
                for g, node in zip(groups, nodes):
                    self.population.add_to_group(g, node)
            else:
                self.population.add_to_group(groups, nodes)

        if n == 1:
            return nodes[0]

        return nodes

    def delete_nodes(self, nodes):
        '''
        Remove nodes (and associated edges) from the graph.
        '''
        self._graph.delete_vertices(nodes)

        for key in self._nattr:
            self._nattr._num_values_set[key] = self.node_nb()

        for key in self._eattr:
            self._eattr._num_values_set[key] = self.edge_nb()

        # check spatial and structure properties
        _post_del_update(self, nodes)

    def new_edge(self, source, target, attributes=None, ignore=False,
                 self_loop=False):
        '''
        Adding a connection to the graph, with optional properties.

        .. versionchanged :: 2.0
            Added `self_loop` argument to enable adding self-loops.

        Parameters
        ----------
        source : :class:`int/node`
            Source node.
        target : :class:`int/node`
            Target node.
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": 1.}``, the unit weight for the
            connection (synaptic strength in NEST).
        ignore : bool, optional (default: False)
            If set to True, ignore attempts to add an existing edge and accept
            self-loops; otherwise an error is raised.
        self_loop : bool, optional (default: False)
            Whether to allow self-loops or not.

        Returns
        -------
        The new connection or None if nothing was added.
        '''
        attributes = {} if attributes is None else deepcopy(attributes)

        if source == target:
            if not ignore and not self_loop:
                raise InvalidArgument("Trying to add a self-loop.")
            elif ignore:
                _log_message(logger, "INFO",
                             "Self-loop on {} ignored.".format(source))

                return None

        return self.new_edges(((source, target),), attributes,
                              check_self_loops=(not ignore and not self_loop),
                              ignore_invalid=ignore)

    def new_edges(self, edge_list, attributes=None, check_duplicates=False,
                  check_self_loops=True, check_existing=True,
                  ignore_invalid=False):
        '''
        Add a list of edges to the graph.

        .. versionchanged:: 2.0
            Can perform all possible checks before adding new edges via the
            ``check_duplicates`` ``check_self_loops``, and ``check_existing``
            arguments.

        Parameters
        ----------
        edge_list : list of 2-tuples or np.array of shape (edge_nb, 2)
            List of the edges that should be added as tuples (source, target)
        attributes : :class:`dict`, optional (default: ``{}``)
            Dictionary containing optional edge properties. If the graph is
            weighted, defaults to ``{"weight": ones}``, where ``ones`` is an
            array the same length as the `edge_list` containing a unit weight
            for each connection (synaptic strength in NEST).
        check_duplicates : bool, optional (default: False)
            Check for duplicate edges within `edge_list`.
        check_self_loops : bool, optional (default: True)
            Check for self-loops.
        check_existing : bool, optional (default: True)
            Check whether some of the edges in `edge_list` already exist in the
            graph or exist multiple times in `edge_list` (also performs
            `check_duplicates`).
        ignore_invalid : bool, optional (default: False)
            Ignore invalid edges: they are not added to the graph and are
            silently dropped. Unless this is set to true, an error is raised
            whenever one of the three checks fails.

        .. warning::

            Setting `check_existing` to False will lead to undefined behavior
            if existing edges are provided! Only use it (for speedup) if you
            are sure that you are indeed only adding new edges.

        Returns
        -------
        Returns new edges only.
        '''
        attributes = {} if attributes is None else deepcopy(attributes)
        num_edges  = len(edge_list)

        # check that all nodes exist
        if num_edges:
            if np.max(edge_list) >= self.node_nb():
                raise InvalidArgument("Some nodes do no exist.")

            # set default values for attributes that were not passed
            _set_default_edge_attributes(self, attributes, num_edges)

            # check edges
            new_attr = None

            if check_duplicates or check_self_loops or check_existing:
                edge_list, new_attr = _cleanup_edges(
                    self, edge_list, attributes, check_duplicates,
                    check_self_loops, check_existing, ignore_invalid)
            else:
                new_attr = attributes

            self._graph.add_edges(edge_list)

            # check distance
            _set_dist_new_edges(new_attr, self, edge_list)

            # call parent function to set the attributes
            self._attr_new_edges(edge_list, attributes=new_attr)

            if len(edge_list) == 0:
                return None
            elif len(edge_list) == 1:
                return edge_list[0]

        return edge_list

    def delete_edges(self, edges):
        ''' Remove a list of edges '''
        if nonstring_container(edges[0]):
            if isinstance(edges[0], tuple):
                self._graph.delete_edges(edges)
            else:
                self._graph.delete_edges([tuple(e) for e in edges])
        else:
            self._graph.delete_edges([edges])

        for key in self._eattr:
            self._eattr._num_values_set[key] = self.edge_nb()

    def clear_all_edges(self):
        ''' Remove all edges from the graph '''
        self._graph.delete_edges(None)
        self._eattr.clear()

    #-------------------------------------------------------------------------#
    # Getters

    def node_nb(self):
        ''' Number of nodes in the graph '''
        return self._graph.vcount()

    def edge_nb(self):
        ''' Number of edges in the graph '''
        return self._graph.ecount()

    def get_degrees(self, mode="total", nodes=None, weights=None):
        g = self._graph
        w = _get_ig_weights(self, weights)
    
        mode = 'all' if mode == 'total' else mode

        if nonstring_container(weights) or weights not in {False, None}:
            return np.array(g.strength(nodes, mode=mode, weights=w))

        return np.array(g.degree(nodes, mode=mode), dtype=int)

    def is_connected(self, mode="strong"):
        '''
        Return whether the graph is connected.

        Parameters
        ----------
        mode : str, optional (default: "strong")
            Whether to test connectedness with directed ("strong") or
            undirected ("weak") connections.
        '''
        return self._graph.is_connected(mode)

    def neighbours(self, node, mode="all"):
        '''
        Return the neighbours of `node`.

        Parameters
        ----------
        node : int
            Index of the node of interest.
        mode : string, optional (default: "all")
            Type of neighbours that will be returned: "all" returns all the
            neighbours regardless of directionality, "in" returns the
            in-neighbours (also called predecessors) and "out" retruns the
            out-neighbours (or successors).

        Returns
        -------
        neighbours : set
            The neighbours of `node`.
        '''
        g = self._graph

        if mode == "all":
            return set(n for n in g.neighbors(node, mode=3))
        elif mode == "in":
            return set(n for n in g.neighbors(node, mode=2))
        elif mode == "out":
            return set(n for n in g.neighbors(node, mode=1))

        raise ArgumentError('Invalid `mode` argument {}; possible values are '
                            '"all", "out" or "in".'.format(mode))

    def _from_library_graph(self, graph, copy=True):
        ''' Initialize `self._graph` from existing library object. '''
        nodes = graph.vcount()
        edges = graph.ecount()

        self._graph = graph.copy() if copy else graph

        # get attributes names and "types" and initialize them
        if nodes:
            for key, val in graph.vs[0].attributes().items():
                super(type(self._nattr), self._nattr).__setitem__(
                    key, _get_dtype(val))

        if edges:
            for key, val in graph.es[0].attributes().items():
                if key != 'eid':
                    super(type(self._eattr), self._eattr).__setitem__(
                        key, _get_dtype(val))


def _get_weights(g, weights):
    if weights in g.edge_attributes:
        # existing edge attribute
        return np.array(g._graph.es[weights])
    elif nonstring_container(weights):
        # user-provided array
        return np.array(weights)
    elif weights is True:
        # "normal" weights
        return np.array(g._graph.es["weight"])
    elif not weights:
        # unweighted
        return None

    raise ValueError("Unknown edge attribute '" + str(weights) + "'.")



# -*- coding: utf-8 -*-
"""
    Extension to save typing and prevent hard-coding of base URLs in the reST
    files.

    This adds a new config value called ``extlinks_fancy`` that is created like this::

       extlinks_fancy = {'exmpl': ('http://example.com/{0}.html', "Example {0}"), ...}

    You can also give an explicit caption, e.g. :exmpl:`Foo <foo>`.

    :copyright: Copyright 2007-2017 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from six import iteritems
from docutils import nodes, utils

import sphinx
from sphinx.util.nodes import split_explicit_title


def make_link_role(base_urls, prefixes):
    def role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
        text = utils.unescape(text)
        has_explicit_title, title, part = split_explicit_title(text)
        pnodes = []
        for base_url, prefix in zip(base_urls, prefixes):
            full_url = None

            if typ == "gtdoc":
                # graph-tool documentation formatting
                module   = part.split(".")[0]
                full_url = base_url.format(module, part)
            elif typ == "nxdoc":
                # networkx documentation formatting
                module   = part.split(".")[0]

                if not module:
                    part = part[1:]
                else:
                    module += "/"

                full_url = base_url.format(module, part)
            else:
                full_url = base_url.format(part)

            if not has_explicit_title:
                if prefix is None:
                    title = full_url
                else:
                    title = prefix.format(part)
            ref = nodes.reference(title, title, internal=False, refuri=full_url)
            if len(pnodes) == 1:
                pnodes.append(nodes.Text(" ["))
            if len(pnodes) > 2:
                pnodes.append(nodes.Text(", "))
            pnodes.append(ref)
        if len(base_urls) > 1:
            pnodes.append(nodes.Text("]"))
        return pnodes, []
    return role


def setup_link_roles(app):
    for name, (base_urls, prefixes) in iteritems(app.config.extlinks_fancy):
        app.add_role(name, make_link_role(base_urls, prefixes))


def setup(app):
    app.add_config_value('extlinks_fancy', {}, 'env')
    app.connect('builder-inited', setup_link_roles)
    return {'version': sphinx.__display_version__, 'parallel_read_safe': True}



#!/usr/bin/env python
#-*- coding:utf-8 -*-
#
# This file is part of the NNGT project to generate and analyze
# neuronal networks and their activity.
# Copyright (C) 2015-2019  Tanguy Fardet
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Spatial graphs generation and methods '''

import os
import time

import numpy as np

import nngt
from nngt.geometry import Shape

# nngt.seed(0)


# ---------------------------- #
# Generate the spatial network #
# ---------------------------- #

ell = Shape.ellipse(radii=(3000., 5000.))

num_nodes  = 1000
population = nngt.NeuralPop.uniform(num_nodes)

g = nngt.generation.gaussian_degree(
    100., 5., nodes=num_nodes, shape=ell, population=population)


# -------------- #
# Saving/loading #
# -------------- #

start = time.time()
g.to_file('sp_graph.el')
print('Saving in {} s.'.format(time.time() - start))

start = time.time()
g2 = nngt.Graph.from_file('sp_graph.el')
print('Loading in {} s.'.format(time.time() - start))

# check equality of shapes and populations

print('Both networks have same area: {}.'.format(
      np.isclose(g2.shape.area, ell.area)))
print('They also have the same boundaries: {}.'.format(
      np.all(np.isclose(g2.shape.bounds, ell.bounds))))

same_groups = np.all(
    [g2.population[k] == g.population[k] for k in g.population])
same_ids = np.all(
    [g2.population[k].ids == g.population[k].ids for k in g.population])

print('They also have the same population: {}.'.format(same_groups * same_ids))


# remove file
os.remove('sp_graph.el')


# ---- #
# Plot #
# ---- #

if nngt.get_config('with_plot'):
    nngt.plot.draw_network(g2, decimate_connections=100, show=True)
